"""
    :codeauthor: Piter Punk <piterpunk@slackware.com>
"""

import ast
from collections import OrderedDict

import pytest

import salt.states.zabbix_host as zabbix_host
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {zabbix_host: {"__opts__": {"test": False}}}


@pytest.fixture()
def basic_host_configuration():
    host = "new_host"
    groups = ["Testing Group"]
    interfaces = [
        OrderedDict(
            [
                (
                    "basic_interface",
                    [
                        OrderedDict([("ip", "127.0.0.1")]),
                        OrderedDict([("type", "agent")]),
                    ],
                )
            ]
        )
    ]
    kwargs = {
        "_connection_user": "XXXXXXXXXX",
        "_connection_password": "XXXXXXXXXX",
        "_connection_url": "http://XXXXXXXXX/zabbix/api_jsonrpc.php",
    }

    ret = {
        "changes": {
            "new_host": {
                "new": "Host new_host created.",
                "old": "Host new_host does not exist.",
            }
        },
        "comment": "Host new_host created.",
        "name": "new_host",
        "result": True,
    }
    return host, groups, interfaces, kwargs, ret


@pytest.fixture()
def existing_host_responses():
    host_get_output = [
        {
            "hostid": "31337",
            "host": "new_host",
            "auto_compress": "1",
            "available": "1",
            "description": "",
            "disable_until": "0",
            "discover": "0",
            "error": "",
            "errors_from": "0",
            "flags": "0",
            "inventory_mode": "-1",
            "ipmi_authtype": "-1",
            "ipmi_available": "0",
            "ipmi_disable_until": "0",
            "ipmi_error": "",
            "ipmi_errors_from": "0",
            "ipmi_password": "",
            "ipmi_privilege": "2",
            "ipmi_username": "",
            "jmx_available": "0",
            "jmx_disable_until": "0",
            "jmx_error": "",
            "jmx_errors_from": "0",
            "lastaccess": "0",
            "maintenance_from": "0",
            "maintenance_status": "0",
            "maintenance_type": "0",
            "maintenanceid": "0",
            "name": "",
            "proxy_address": "",
            "proxy_hostid": "0",
            "snmp_available": "0",
            "snmp_disable_until": "0",
            "snmp_error": "",
            "snmp_errors_from": "0",
            "status": "0",
            "templateid": "0",
            "tls_accept": "1",
            "tls_connect": "1",
            "tls_issuer": "",
            "tls_psk": "",
            "tls_psk_identity": "",
            "tls_subject": "",
        }
    ]
    hostgroup_get_output_up = [
        {"groupid": "16", "name": "Testing Group", "internal": "0", "flags": "0"}
    ]
    hostinterface_get_output = [
        {
            "interfaceid": "29",
            "hostid": "31337",
            "main": "1",
            "type": "1",
            "useip": "1",
            "ip": "127.0.0.1",
            "dns": "basic_interface",
            "port": "10050",
            "details": [],
        }
    ]
    host_inventory_get_output = False

    return (
        host_get_output,
        hostgroup_get_output_up,
        hostinterface_get_output,
        host_inventory_get_output,
    )


def test_create_a_basic_new_host(basic_host_configuration):
    """
    This test creates a host with the minimum required parameters:

    host, groups, interfaces and _connection_args

    The "groups" should be converted to their numeric IDs and the
    hidden mandatory fields of interfaces should be populated.
    """
    host, groups, interfaces, kwargs, ret = basic_host_configuration

    hostgroup_get_output = [
        {"groupid": "16", "name": "Testing Group", "internal": "0", "flags": "0"}
    ]
    host_exists_output = False
    host_create_output = "31337"

    mock_hostgroup_get = MagicMock(return_value=hostgroup_get_output)
    mock_host_exists = MagicMock(return_value=host_exists_output)
    mock_host_create = MagicMock(return_value=host_create_output)
    with patch.dict(
        zabbix_host.__salt__,
        {
            "zabbix.hostgroup_get": mock_hostgroup_get,
            "zabbix.host_exists": mock_host_exists,
            "zabbix.host_create": mock_host_create,
        },
    ):
        assert zabbix_host.present(host, groups, interfaces, **kwargs) == ret
        mock_host_create.assert_called_with(
            "new_host",
            [16],
            [
                {
                    "type": "1",
                    "main": "1",
                    "useip": "1",
                    "ip": "127.0.0.1",
                    "dns": "basic_interface",
                    "port": "10050",
                    "details": [],
                }
            ],
            _connection_password="XXXXXXXXXX",
            _connection_url="http://XXXXXXXXX/zabbix/api_jsonrpc.php",
            _connection_user="XXXXXXXXXX",
            inventory={},
            proxy_hostid="0",
            visible_name=None,
        )


def test_create_a_new_host_with_multiple_groups(basic_host_configuration):
    """
    This test creates a host with multiple groups, mixing names and IDs.
    """
    host, _, interfaces, kwargs, ret = basic_host_configuration
    groups = ["Testing Group", 15, "Tested Group"]

    hostgroup_get_output = [
        [{"groupid": "16", "name": "Testing Group", "internal": "0", "flags": "0"}],
        [{"groupid": "17", "name": "Tested Group", "internal": "0", "flags": "0"}],
    ]
    host_exists_output = False
    host_create_output = "31337"

    mock_hostgroup_get = MagicMock(side_effect=hostgroup_get_output)
    mock_host_exists = MagicMock(return_value=host_exists_output)
    mock_host_create = MagicMock(return_value=host_create_output)
    with patch.dict(
        zabbix_host.__salt__,
        {
            "zabbix.hostgroup_get": mock_hostgroup_get,
            "zabbix.host_exists": mock_host_exists,
            "zabbix.host_create": mock_host_create,
        },
    ):
        assert zabbix_host.present(host, groups, interfaces, **kwargs) == ret
        mock_host_create.assert_called_with(
            "new_host",
            [16, 15, 17],
            [
                {
                    "type": "1",
                    "main": "1",
                    "useip": "1",
                    "ip": "127.0.0.1",
                    "dns": "basic_interface",
                    "port": "10050",
                    "details": [],
                }
            ],
            _connection_password="XXXXXXXXXX",
            _connection_url="http://XXXXXXXXX/zabbix/api_jsonrpc.php",
            _connection_user="XXXXXXXXXX",
            inventory={},
            proxy_hostid="0",
            visible_name=None,
        )


def test_create_a_new_host_with_multiple_interfaces(basic_host_configuration):
    """
    Tests the creation of a host with multiple interfaces. This creates
    one interface of each type, which needs to have their default
    parameters filled. Also, tests the different dns, ip and useip
    combinations.
    """
    host, groups, _, kwargs, ret = basic_host_configuration
    interfaces = [
        OrderedDict(
            [
                (
                    "agent_interface",
                    [
                        OrderedDict([("dns", "new_host")]),
                        OrderedDict([("type", "agent")]),
                    ],
                ),
                (
                    "snmp_interface",
                    [
                        OrderedDict([("ip", "127.0.0.1")]),
                        OrderedDict([("dns", "new_host")]),
                        OrderedDict([("useip", False)]),
                        OrderedDict([("type", "snmp")]),
                    ],
                ),
                (
                    "ipmi_interface",
                    [
                        OrderedDict([("ip", "127.0.0.1")]),
                        OrderedDict([("dns", "new_host")]),
                        OrderedDict([("type", "ipmi")]),
                    ],
                ),
                (
                    "jmx_interface",
                    [
                        OrderedDict([("ip", "127.0.0.1")]),
                        OrderedDict([("dns", "new_host")]),
                        OrderedDict([("useip", True)]),
                        OrderedDict([("type", "jmx")]),
                    ],
                ),
            ]
        )
    ]

    hostgroup_get_output = [
        [{"groupid": "16", "name": "Testing Group", "internal": "0", "flags": "0"}],
        [{"groupid": "17", "name": "Tested Group", "internal": "0", "flags": "0"}],
    ]
    host_exists_output = False
    host_create_output = "31337"

    mock_hostgroup_get = MagicMock(side_effect=hostgroup_get_output)
    mock_host_exists = MagicMock(return_value=host_exists_output)
    mock_host_create = MagicMock(return_value=host_create_output)
    with patch.dict(
        zabbix_host.__salt__,
        {
            "zabbix.hostgroup_get": mock_hostgroup_get,
            "zabbix.host_exists": mock_host_exists,
            "zabbix.host_create": mock_host_create,
        },
    ):
        assert zabbix_host.present(host, groups, interfaces, **kwargs) == ret
        # Blame Python 3.5 for this:
        host_create_call = mock_host_create.call_args[0]
        assert host_create_call[0] == "new_host"
        assert host_create_call[1] == [16]
        for interface in host_create_call[2]:
            if interface["type"] == "1":
                assert interface == {
                    "type": "1",
                    "main": "1",
                    "useip": "1",
                    "ip": "",
                    "dns": "new_host",
                    "port": "10050",
                    "details": [],
                }
            elif interface["type"] == "2":
                assert interface == {
                    "type": "2",
                    "main": "1",
                    "useip": "0",
                    "ip": "127.0.0.1",
                    "dns": "new_host",
                    "port": "161",
                    "details": {
                        "version": "2",
                        "bulk": "1",
                        "community": "{$SNMP_COMMUNITY}",
                    },
                }
            elif interface["type"] == "3":
                assert interface == {
                    "type": "3",
                    "main": "1",
                    "useip": "1",
                    "ip": "127.0.0.1",
                    "dns": "new_host",
                    "port": "623",
                    "details": [],
                }
            elif interface["type"] == "4":
                assert interface == {
                    "type": "4",
                    "main": "1",
                    "useip": "1",
                    "ip": "127.0.0.1",
                    "dns": "new_host",
                    "port": "12345",
                    "details": [],
                }
            else:
                assert interface["type"] == "Should be 1, 2, 3 or 4"


def test_create_a_new_host_with_additional_parameters(basic_host_configuration):
    """
    Tests if additional parameters, like "description" or "inventory_mode"
    are being properly passed to host_create. Also, checks if invalid
    parameters are filtered out.
    """
    host, groups, interfaces, kwargs, ret = basic_host_configuration
    kwargs["visible_name"] = "Visible Name"
    kwargs["description"] = "An amazing test host entry"
    kwargs["not_valid_property"] = "This should be removed"
    kwargs["inventory_mode"] = "0"

    hostgroup_get_output = [
        {"groupid": "16", "name": "Testing Group", "internal": "0", "flags": "0"}
    ]
    host_exists_output = False
    host_create_output = "31337"

    mock_hostgroup_get = MagicMock(return_value=hostgroup_get_output)
    mock_host_exists = MagicMock(return_value=host_exists_output)
    mock_host_create = MagicMock(return_value=host_create_output)
    with patch.dict(
        zabbix_host.__salt__,
        {
            "zabbix.hostgroup_get": mock_hostgroup_get,
            "zabbix.host_exists": mock_host_exists,
            "zabbix.host_create": mock_host_create,
        },
    ):
        assert zabbix_host.present(host, groups, interfaces, **kwargs) == ret
        mock_host_create.assert_called_with(
            "new_host",
            [16],
            [
                {
                    "type": "1",
                    "main": "1",
                    "useip": "1",
                    "ip": "127.0.0.1",
                    "dns": "basic_interface",
                    "port": "10050",
                    "details": [],
                }
            ],
            _connection_password="XXXXXXXXXX",
            _connection_url="http://XXXXXXXXX/zabbix/api_jsonrpc.php",
            _connection_user="XXXXXXXXXX",
            description="An amazing test host entry",
            inventory={},
            inventory_mode="0",
            proxy_hostid="0",
            visible_name="Visible Name",
        )


def test_create_a_new_host_with_proxy_by_name(basic_host_configuration):
    """
    Test the handling of proxy_host parameter when it is a name
    """
    host, groups, interfaces, kwargs, ret = basic_host_configuration
    kwargs["proxy_host"] = "RemoteProxy"

    hostgroup_get_output = [
        {"groupid": "16", "name": "Testing Group", "internal": "0", "flags": "0"}
    ]
    host_exists_output = False
    host_create_output = "31337"
    run_query_output = [
        {
            "proxyid": "10356",
            "interface": {
                "interfaceid": "56",
                "hostid": "10356",
                "main": "1",
                "type": "0",
                "useip": "1",
                "ip": "127.0.0.1",
                "dns": "remoteproxy.veryfar",
                "port": "10051",
                "details": [],
            },
        }
    ]

    mock_hostgroup_get = MagicMock(return_value=hostgroup_get_output)
    mock_host_exists = MagicMock(return_value=host_exists_output)
    mock_host_create = MagicMock(return_value=host_create_output)
    mock_run_query = MagicMock(return_value=run_query_output)
    with patch.dict(
        zabbix_host.__salt__,
        {
            "zabbix.hostgroup_get": mock_hostgroup_get,
            "zabbix.host_exists": mock_host_exists,
            "zabbix.host_create": mock_host_create,
            "zabbix.run_query": mock_run_query,
        },
    ):
        assert zabbix_host.present(host, groups, interfaces, **kwargs) == ret
        mock_host_create.assert_called_with(
            "new_host",
            [16],
            [
                {
                    "type": "1",
                    "main": "1",
                    "useip": "1",
                    "ip": "127.0.0.1",
                    "dns": "basic_interface",
                    "port": "10050",
                    "details": [],
                }
            ],
            _connection_password="XXXXXXXXXX",
            _connection_url="http://XXXXXXXXX/zabbix/api_jsonrpc.php",
            _connection_user="XXXXXXXXXX",
            inventory={},
            proxy_hostid="10356",
            visible_name=None,
        )


def test_create_a_new_host_with_proxy_by_id(basic_host_configuration):
    """
    Test the handling of proxy_host parameter when it is a proxyid
    """
    host, groups, interfaces, kwargs, ret = basic_host_configuration
    kwargs["proxy_host"] = 10356

    hostgroup_get_output = [
        {"groupid": "16", "name": "Testing Group", "internal": "0", "flags": "0"}
    ]
    host_exists_output = False
    host_create_output = "31337"
    run_query_output = [{"proxyid": "10356"}]

    mock_hostgroup_get = MagicMock(return_value=hostgroup_get_output)
    mock_host_exists = MagicMock(return_value=host_exists_output)
    mock_host_create = MagicMock(return_value=host_create_output)
    mock_run_query = MagicMock(return_value=run_query_output)
    with patch.dict(
        zabbix_host.__salt__,
        {
            "zabbix.hostgroup_get": mock_hostgroup_get,
            "zabbix.host_exists": mock_host_exists,
            "zabbix.host_create": mock_host_create,
            "zabbix.run_query": mock_run_query,
        },
    ):
        assert zabbix_host.present(host, groups, interfaces, **kwargs) == ret
        mock_host_create.assert_called_with(
            "new_host",
            [16],
            [
                {
                    "type": "1",
                    "main": "1",
                    "useip": "1",
                    "ip": "127.0.0.1",
                    "dns": "basic_interface",
                    "port": "10050",
                    "details": [],
                }
            ],
            _connection_password="XXXXXXXXXX",
            _connection_url="http://XXXXXXXXX/zabbix/api_jsonrpc.php",
            _connection_user="XXXXXXXXXX",
            inventory={},
            proxy_hostid="10356",
            visible_name=None,
        )


def test_create_a_new_host_with_missing_groups(basic_host_configuration):
    """
    Tests when any of the provided groups doesn't exists
    """
    host, _, interfaces, kwargs, _ = basic_host_configuration
    groups = ["Testing Group", "Missing Group"]

    ret = {
        "changes": {},
        "comment": "Invalid group Missing Group",
        "name": "new_host",
        "result": False,
    }

    hostgroup_get_output = [
        [{"groupid": "16", "name": "Testing Group", "internal": "0", "flags": "0"}],
        False,
    ]
    host_exists_output = False
    host_create_output = "31337"

    mock_hostgroup_get = MagicMock(side_effect=hostgroup_get_output)
    mock_host_exists = MagicMock(return_value=host_exists_output)
    mock_host_create = MagicMock(return_value=host_create_output)
    with patch.dict(
        zabbix_host.__salt__,
        {
            "zabbix.hostgroup_get": mock_hostgroup_get,
            "zabbix.host_exists": mock_host_exists,
            "zabbix.host_create": mock_host_create,
        },
    ):
        assert zabbix_host.present(host, groups, interfaces, **kwargs) == ret
        assert not mock_host_create.called, "host_create should not be called"


def test_create_a_new_host_with_missing_proxy(basic_host_configuration):
    """
    Tests when the given proxy_host doesn't exists
    """
    host, groups, interfaces, kwargs, _ = basic_host_configuration
    kwargs["proxy_host"] = 10356

    ret = {
        "changes": {},
        "comment": "Invalid proxy_host 10356",
        "name": "new_host",
        "result": False,
    }

    hostgroup_get_output = [
        {"groupid": "16", "name": "Testing Group", "internal": "0", "flags": "0"}
    ]
    host_exists_output = False
    host_create_output = "31337"
    run_query_output = False

    mock_hostgroup_get = MagicMock(return_value=hostgroup_get_output)
    mock_host_exists = MagicMock(return_value=host_exists_output)
    mock_run_query = MagicMock(return_value=run_query_output)
    mock_host_create = MagicMock(return_value=host_create_output)
    with patch.dict(
        zabbix_host.__salt__,
        {
            "zabbix.hostgroup_get": mock_hostgroup_get,
            "zabbix.host_exists": mock_host_exists,
            "zabbix.run_query": mock_run_query,
            "zabbix.host_create": mock_host_create,
        },
    ):
        assert zabbix_host.present(host, groups, interfaces, **kwargs) == ret
        assert not mock_host_create.called, "host_create should not be called"


def test_create_a_new_host_with_inventory_as_a_list(basic_host_configuration):
    """
    This test creates a host with a populated inventory declared as a list
    """
    host, groups, interfaces, kwargs, ret = basic_host_configuration
    inventory = (
        {"vendor": "FakeVendor"},
        {"asset_tag": "ABC12345"},
    )

    hostgroup_get_output = [
        {"groupid": "16", "name": "Testing Group", "internal": "0", "flags": "0"}
    ]
    host_exists_output = False
    host_create_output = "31337"

    mock_hostgroup_get = MagicMock(return_value=hostgroup_get_output)
    mock_host_exists = MagicMock(return_value=host_exists_output)
    mock_host_create = MagicMock(return_value=host_create_output)
    with patch.dict(
        zabbix_host.__salt__,
        {
            "zabbix.hostgroup_get": mock_hostgroup_get,
            "zabbix.host_exists": mock_host_exists,
            "zabbix.host_create": mock_host_create,
        },
    ):
        assert (
            zabbix_host.present(host, groups, interfaces, inventory=inventory, **kwargs)
            == ret
        )
        mock_host_create.assert_called_with(
            "new_host",
            [16],
            [
                {
                    "type": "1",
                    "main": "1",
                    "useip": "1",
                    "ip": "127.0.0.1",
                    "dns": "basic_interface",
                    "port": "10050",
                    "details": [],
                }
            ],
            _connection_password="XXXXXXXXXX",
            _connection_url="http://XXXXXXXXX/zabbix/api_jsonrpc.php",
            _connection_user="XXXXXXXXXX",
            inventory={"vendor": "FakeVendor", "asset_tag": "ABC12345"},
            proxy_hostid="0",
            visible_name=None,
        )


def test_create_a_new_host_with_inventory_as_a_dict(basic_host_configuration):
    """
    This test creates a host with a populated inventory declared as a dictionary
    """
    host, groups, interfaces, kwargs, ret = basic_host_configuration
    inventory = {
        "vendor": "FakeVendor",
        "asset_tag": "ABC12345",
    }

    hostgroup_get_output = [
        {"groupid": "16", "name": "Testing Group", "internal": "0", "flags": "0"}
    ]
    host_exists_output = False
    host_create_output = "31337"

    mock_hostgroup_get = MagicMock(return_value=hostgroup_get_output)
    mock_host_exists = MagicMock(return_value=host_exists_output)
    mock_host_create = MagicMock(return_value=host_create_output)
    with patch.dict(
        zabbix_host.__salt__,
        {
            "zabbix.hostgroup_get": mock_hostgroup_get,
            "zabbix.host_exists": mock_host_exists,
            "zabbix.host_create": mock_host_create,
        },
    ):
        assert (
            zabbix_host.present(host, groups, interfaces, inventory=inventory, **kwargs)
            == ret
        )
        mock_host_create.assert_called_with(
            "new_host",
            [16],
            [
                {
                    "type": "1",
                    "main": "1",
                    "useip": "1",
                    "ip": "127.0.0.1",
                    "dns": "basic_interface",
                    "port": "10050",
                    "details": [],
                }
            ],
            _connection_password="XXXXXXXXXX",
            _connection_url="http://XXXXXXXXX/zabbix/api_jsonrpc.php",
            _connection_user="XXXXXXXXXX",
            inventory={"vendor": "FakeVendor", "asset_tag": "ABC12345"},
            proxy_hostid="0",
            visible_name=None,
        )


def test_ensure_nothing_happens_when_host_is_in_desired_state(
    basic_host_configuration, existing_host_responses
):
    """
    Test to ensure that nothing happens when the state applied
    already corresponds to the host actual configuration.
    """
    host, groups, interfaces, kwargs, _ = basic_host_configuration
    (
        host_get_output,
        hostgroup_get_output_up,
        hostinterface_get_output,
        host_inventory_get_output,
    ) = existing_host_responses

    hostgroup_get_output = [
        [{"groupid": "16", "name": "Testing Group", "internal": "0", "flags": "0"}],
        hostgroup_get_output_up,
    ]
    host_exists_output = True
    host_create_output = "31337"

    ret = {
        "changes": {},
        "comment": "Host new_host already exists.",
        "name": "new_host",
        "result": True,
    }

    mock_hostgroup_get = MagicMock(side_effect=hostgroup_get_output)
    mock_host_exists = MagicMock(return_value=host_exists_output)
    mock_host_get = MagicMock(return_value=host_get_output)
    mock_hostinterface_get = MagicMock(return_value=hostinterface_get_output)
    mock_host_inventory_get = MagicMock(return_value=host_inventory_get_output)
    mock_host_update = MagicMock(return_value=False)
    with patch.dict(
        zabbix_host.__salt__,
        {
            "zabbix.hostgroup_get": mock_hostgroup_get,
            "zabbix.host_exists": mock_host_exists,
            "zabbix.host_get": mock_host_get,
            "zabbix.hostinterface_get": mock_hostinterface_get,
            "zabbix.host_inventory_get": mock_host_inventory_get,
            "zabbix.host_update": mock_host_update,
        },
    ):
        assert zabbix_host.present(host, groups, interfaces, **kwargs) == ret
        assert not mock_host_update.called, "host_update should not be called"


def test_change_a_host_group(basic_host_configuration, existing_host_responses):
    """
    Tests if the group of a host is changed when solicited
    """
    host, groups, interfaces, kwargs, _ = basic_host_configuration
    (
        host_get_output,
        hostgroup_get_output_up,
        hostinterface_get_output,
        host_inventory_get_output,
    ) = existing_host_responses

    hostgroup_get_output = [
        hostgroup_get_output_up,
        [{"groupid": "17", "name": "Actual Group", "internal": "0", "flags": "0"}],
    ]
    host_exists_output = True
    host_update_output = "31337"

    ret = {
        "changes": {"groups": "[16]"},
        "comment": "Host new_host updated.",
        "name": "new_host",
        "result": True,
    }

    mock_hostgroup_get = MagicMock(side_effect=hostgroup_get_output)
    mock_host_exists = MagicMock(return_value=host_exists_output)
    mock_host_get = MagicMock(return_value=host_get_output)
    mock_hostinterface_get = MagicMock(return_value=hostinterface_get_output)
    mock_host_inventory_get = MagicMock(return_value=host_inventory_get_output)
    mock_host_update = MagicMock(return_value=host_update_output)
    with patch.dict(
        zabbix_host.__salt__,
        {
            "zabbix.hostgroup_get": mock_hostgroup_get,
            "zabbix.host_exists": mock_host_exists,
            "zabbix.host_get": mock_host_get,
            "zabbix.hostinterface_get": mock_hostinterface_get,
            "zabbix.host_inventory_get": mock_host_inventory_get,
            "zabbix.host_update": mock_host_update,
        },
    ):
        assert zabbix_host.present(host, groups, interfaces, **kwargs) == ret
        mock_host_update.assert_called_with(
            "31337",
            groups=[16],
            _connection_password="XXXXXXXXXX",
            _connection_url="http://XXXXXXXXX/zabbix/api_jsonrpc.php",
            _connection_user="XXXXXXXXXX",
        )


def test_to_add_new_groups_to_a_host(basic_host_configuration, existing_host_responses):
    """
    Tests if new groups are added to a host
    """
    host, _, interfaces, kwargs, _ = basic_host_configuration
    (
        host_get_output,
        hostgroup_get_output_up,
        hostinterface_get_output,
        host_inventory_get_output,
    ) = existing_host_responses

    groups = ["Testing Group", 15, "Tested Group"]

    hostgroup_get_output = [
        hostgroup_get_output_up,
        [{"groupid": "17", "name": "Actual Group", "internal": "0", "flags": "0"}],
        hostgroup_get_output_up,
    ]
    host_exists_output = True
    host_update_output = "31337"

    ret = {
        "changes": {"groups": "[16, 15, 17]"},
        "comment": "Host new_host updated.",
        "name": "new_host",
        "result": True,
    }

    mock_hostgroup_get = MagicMock(side_effect=hostgroup_get_output)
    mock_host_exists = MagicMock(return_value=host_exists_output)
    mock_host_get = MagicMock(return_value=host_get_output)
    mock_hostinterface_get = MagicMock(return_value=hostinterface_get_output)
    mock_host_inventory_get = MagicMock(return_value=host_inventory_get_output)
    mock_host_update = MagicMock(return_value=host_update_output)
    with patch.dict(
        zabbix_host.__salt__,
        {
            "zabbix.hostgroup_get": mock_hostgroup_get,
            "zabbix.host_exists": mock_host_exists,
            "zabbix.host_get": mock_host_get,
            "zabbix.hostinterface_get": mock_hostinterface_get,
            "zabbix.host_inventory_get": mock_host_inventory_get,
            "zabbix.host_update": mock_host_update,
        },
    ):
        assert zabbix_host.present(host, groups, interfaces, **kwargs) == ret
        mock_host_update.assert_called_with(
            "31337",
            groups=[16, 15, 17],
            _connection_password="XXXXXXXXXX",
            _connection_url="http://XXXXXXXXX/zabbix/api_jsonrpc.php",
            _connection_user="XXXXXXXXXX",
        )


def test_update_an_existent_host_proxy(
    basic_host_configuration, existing_host_responses
):
    """
    Tests if the proxy of a host is updated to a new one.
    This also tests if a proxy can be added, as a host without a proxy
    have the proxy_hostid property equals zero.
    """
    host, groups, interfaces, kwargs, _ = basic_host_configuration
    (
        host_get_output,
        hostgroup_get_output,
        hostinterface_get_output,
        host_inventory_get_output,
    ) = existing_host_responses

    kwargs["proxy_host"] = 10356
    host_exists_output = True
    host_update_output = "31337"
    run_query_output = [{"proxyid": "10356"}]

    ret = {
        "changes": {"proxy_hostid": "10356"},
        "comment": "Host new_host updated.",
        "name": "new_host",
        "result": True,
    }

    mock_hostgroup_get = MagicMock(return_value=hostgroup_get_output)
    mock_host_exists = MagicMock(return_value=host_exists_output)
    mock_host_get = MagicMock(return_value=host_get_output)
    mock_hostinterface_get = MagicMock(return_value=hostinterface_get_output)
    mock_host_inventory_get = MagicMock(return_value=host_inventory_get_output)
    mock_host_update = MagicMock(return_value=host_update_output)
    mock_run_query = MagicMock(return_value=run_query_output)
    with patch.dict(
        zabbix_host.__salt__,
        {
            "zabbix.hostgroup_get": mock_hostgroup_get,
            "zabbix.host_exists": mock_host_exists,
            "zabbix.host_get": mock_host_get,
            "zabbix.hostinterface_get": mock_hostinterface_get,
            "zabbix.host_inventory_get": mock_host_inventory_get,
            "zabbix.host_update": mock_host_update,
            "zabbix.run_query": mock_run_query,
        },
    ):
        assert zabbix_host.present(host, groups, interfaces, **kwargs) == ret
        mock_host_update.assert_called_with(
            "31337",
            proxy_hostid="10356",
            _connection_password="XXXXXXXXXX",
            _connection_url="http://XXXXXXXXX/zabbix/api_jsonrpc.php",
            _connection_user="XXXXXXXXXX",
        )


def test_update_a_host_with_additional_parameters(
    basic_host_configuration, existing_host_responses
):
    """
    This test checks if additional parameters can be added to an
    existing host
    """
    host, groups, interfaces, kwargs, _ = basic_host_configuration
    (
        host_get_output,
        hostgroup_get_output,
        hostinterface_get_output,
        host_inventory_get_output,
    ) = existing_host_responses

    kwargs["inventory_mode"] = 0
    kwargs["description"] = "An amazing test host entry"
    host_exists_output = True
    host_update_output = "31337"

    ret = {
        "changes": {
            "host": "{'description': 'An amazing test host entry', 'inventory_mode': 0}"
        },
        "comment": "Host new_host updated.",
        "name": "new_host",
        "result": True,
    }

    mock_hostgroup_get = MagicMock(return_value=hostgroup_get_output)
    mock_host_exists = MagicMock(return_value=host_exists_output)
    mock_host_get = MagicMock(return_value=host_get_output)
    mock_hostinterface_get = MagicMock(return_value=hostinterface_get_output)
    mock_host_inventory_get = MagicMock(return_value=host_inventory_get_output)
    mock_host_update = MagicMock(return_value=host_update_output)
    with patch.dict(
        zabbix_host.__salt__,
        {
            "zabbix.hostgroup_get": mock_hostgroup_get,
            "zabbix.host_exists": mock_host_exists,
            "zabbix.host_get": mock_host_get,
            "zabbix.hostinterface_get": mock_hostinterface_get,
            "zabbix.host_inventory_get": mock_host_inventory_get,
            "zabbix.host_update": mock_host_update,
        },
    ):
        # Blame Python 3.5 support for all this black magic
        host_present_ret = zabbix_host.present(host, groups, interfaces, **kwargs)
        host_present_changes = ast.literal_eval(host_present_ret["changes"]["host"])
        assert host_present_changes == ast.literal_eval(ret["changes"]["host"])
        assert host_present_ret["comment"] == "Host new_host updated."
        assert host_present_ret["name"] == "new_host"
        assert host_present_ret["result"] is True
        # When Python 3.5 is gone, the following line does the job:
        # assert zabbix_host.present(host, groups, interfaces, **kwargs) == ret
        mock_host_update.assert_called_with(
            "31337",
            _connection_password="XXXXXXXXXX",
            _connection_url="http://XXXXXXXXX/zabbix/api_jsonrpc.php",
            _connection_user="XXXXXXXXXX",
            description="An amazing test host entry",
            inventory_mode=0,
        )


def test_update_a_hostinterface(basic_host_configuration, existing_host_responses):
    """
    Tests the update of a current hostinterface of a host.
    """
    host, groups, _, kwargs, _ = basic_host_configuration
    (
        host_get_output,
        hostgroup_get_output,
        hostinterface_get_output,
        host_inventory_get_output,
    ) = existing_host_responses

    interfaces = [
        OrderedDict(
            [
                (
                    "basic_interface",
                    [
                        OrderedDict([("dns", "new_host")]),
                        OrderedDict([("type", "agent")]),
                        OrderedDict([("useip", False)]),
                    ],
                ),
            ]
        )
    ]
    host_exists_output = True
    hostinterface_update_output = "29"

    ret = {
        "changes": {
            "interfaces": (
                "[{'type': '1', 'main': '1', 'useip': '0', 'ip': '', 'dns': 'new_host',"
                " 'port': '10050', 'details': []}]"
            )
        },
        "comment": "Host new_host updated.",
        "name": "new_host",
        "result": True,
    }

    mock_hostgroup_get = MagicMock(return_value=hostgroup_get_output)
    mock_host_exists = MagicMock(return_value=host_exists_output)
    mock_host_get = MagicMock(return_value=host_get_output)
    mock_hostinterface_get = MagicMock(return_value=hostinterface_get_output)
    mock_host_inventory_get = MagicMock(return_value=host_inventory_get_output)
    mock_hostinterface_update = MagicMock(return_value=hostinterface_update_output)
    with patch.dict(
        zabbix_host.__salt__,
        {
            "zabbix.hostgroup_get": mock_hostgroup_get,
            "zabbix.host_exists": mock_host_exists,
            "zabbix.host_get": mock_host_get,
            "zabbix.hostinterface_get": mock_hostinterface_get,
            "zabbix.host_inventory_get": mock_host_inventory_get,
            "zabbix.hostinterface_update": mock_hostinterface_update,
        },
    ):
        # Blame Python 3.5 support for all this black magic
        host_present_ret = zabbix_host.present(host, groups, interfaces, **kwargs)
        host_present_changes = ast.literal_eval(
            host_present_ret["changes"]["interfaces"]
        )
        assert host_present_changes == ast.literal_eval(ret["changes"]["interfaces"])
        assert host_present_ret["comment"] == "Host new_host updated."
        assert host_present_ret["name"] == "new_host"
        assert host_present_ret["result"] is True
        # When Python 3.5 is gone, the following line does the job:
        # assert zabbix_host.present(host, groups, interfaces, **kwargs) == ret
        mock_hostinterface_update.assert_called_with(
            interfaceid="29",
            ip="",
            dns="new_host",
            useip="0",
            type="1",
            main="1",
            port="10050",
            details=[],
            _connection_password="XXXXXXXXXX",
            _connection_url="http://XXXXXXXXX/zabbix/api_jsonrpc.php",
            _connection_user="XXXXXXXXXX",
        )


def test_add_a_new_hostinterface(basic_host_configuration, existing_host_responses):
    """
    Tests the update of a current and creation of a new hostinterface
    of a host.
    """
    host, groups, _, kwargs, _ = basic_host_configuration
    (
        host_get_output,
        hostgroup_get_output,
        hostinterface_get_output,
        host_inventory_get_output,
    ) = existing_host_responses

    interfaces = [
        OrderedDict(
            [
                (
                    "basic_interface",
                    [
                        OrderedDict([("dns", "new_host")]),
                        OrderedDict([("type", "agent")]),
                        OrderedDict([("useip", "0")]),
                    ],
                ),
                (
                    "ipmi_interface",
                    [
                        OrderedDict([("ip", "127.0.0.1")]),
                        OrderedDict([("dns", "new_host")]),
                        OrderedDict([("useip", False)]),
                        OrderedDict([("type", "ipmi")]),
                    ],
                ),
            ]
        )
    ]
    host_exists_output = True
    hostinterface_update_output = "29"
    hostinterface_create_output = "30"

    ret = {
        "changes": {
            "interfaces": (
                "[{'type': '1', 'main': '1', 'useip': '0', 'ip': '', 'dns': 'new_host',"
                " 'port': '10050', 'details': []}, {'type': '3', 'main': '1', 'useip':"
                " '0', 'ip': '127.0.0.1', 'dns': 'new_host', 'port': '623', 'details':"
                " []}]"
            )
        },
        "comment": "Host new_host updated.",
        "name": "new_host",
        "result": True,
    }

    mock_hostgroup_get = MagicMock(return_value=hostgroup_get_output)
    mock_host_exists = MagicMock(return_value=host_exists_output)
    mock_host_get = MagicMock(return_value=host_get_output)
    mock_hostinterface_get = MagicMock(return_value=hostinterface_get_output)
    mock_host_inventory_get = MagicMock(return_value=host_inventory_get_output)
    mock_hostinterface_update = MagicMock(return_value=hostinterface_update_output)
    mock_hostinterface_create = MagicMock(return_value=hostinterface_create_output)
    with patch.dict(
        zabbix_host.__salt__,
        {
            "zabbix.hostgroup_get": mock_hostgroup_get,
            "zabbix.host_exists": mock_host_exists,
            "zabbix.host_get": mock_host_get,
            "zabbix.hostinterface_get": mock_hostinterface_get,
            "zabbix.host_inventory_get": mock_host_inventory_get,
            "zabbix.hostinterface_update": mock_hostinterface_update,
            "zabbix.hostinterface_create": mock_hostinterface_create,
        },
    ):
        # Blame Python 3.5 support for all this black magic
        host_present_ret = zabbix_host.present(host, groups, interfaces, **kwargs)
        for interface in ast.literal_eval(host_present_ret["changes"]["interfaces"]):
            if interface["type"] == "1":
                assert interface == ast.literal_eval(ret["changes"]["interfaces"])[0]
            elif interface["type"] == "3":
                assert interface == ast.literal_eval(ret["changes"]["interfaces"])[1]
            else:
                assert interface["type"] == "Should be 1 or 3"
        assert host_present_ret["comment"] == "Host new_host updated."
        assert host_present_ret["name"] == "new_host"
        assert host_present_ret["result"] is True
        # When Python 3.5 is gone, the following line does the job:
        # assert zabbix_host.present(host, groups, interfaces, **kwargs) == ret
        mock_hostinterface_update.assert_called_with(
            interfaceid="29",
            ip="",
            dns="new_host",
            useip="0",
            type="1",
            main="1",
            port="10050",
            details=[],
            _connection_password="XXXXXXXXXX",
            _connection_url="http://XXXXXXXXX/zabbix/api_jsonrpc.php",
            _connection_user="XXXXXXXXXX",
        )
        mock_hostinterface_create.assert_called_with(
            "31337",
            "127.0.0.1",
            dns="new_host",
            useip="0",
            if_type="3",
            main="1",
            port="623",
            details=[],
            _connection_password="XXXXXXXXXX",
            _connection_url="http://XXXXXXXXX/zabbix/api_jsonrpc.php",
            _connection_user="XXXXXXXXXX",
        )


def test_update_inventory_values(basic_host_configuration, existing_host_responses):
    """
    Tests the update of an inventory value
    """
    host, groups, interfaces, kwargs, _ = basic_host_configuration
    (
        host_get_output,
        hostgroup_get_output,
        hostinterface_get_output,
        host_inventory_get_output,
    ) = existing_host_responses

    host_inventory_get_output = {
        "vendor": "FakeVendor",
        "asset_tag": "ABC12345",
    }
    hostgroup_get_output = [
        {"groupid": "16", "name": "Testing Group", "internal": "0", "flags": "0"}
    ]
    host_exists_output = True
    host_inventory_set_output = {"result": {"hostids": ["31337"]}}

    inventory = (
        {"vendor": "TrueVendor"},
        {"asset_tag": "ABC12345"},
    )
    ret = {
        "changes": {"inventory": "{'vendor': 'TrueVendor', 'asset_tag': 'ABC12345'}"},
        "comment": "Host new_host updated.",
        "name": "new_host",
        "result": True,
    }

    mock_hostgroup_get = MagicMock(return_value=hostgroup_get_output)
    mock_host_exists = MagicMock(return_value=host_exists_output)
    mock_host_get = MagicMock(return_value=host_get_output)
    mock_hostinterface_get = MagicMock(return_value=hostinterface_get_output)
    mock_host_inventory_get = MagicMock(return_value=host_inventory_get_output)
    mock_host_inventory_set = MagicMock(return_value=host_inventory_set_output)
    with patch.dict(
        zabbix_host.__salt__,
        {
            "zabbix.hostgroup_get": mock_hostgroup_get,
            "zabbix.host_exists": mock_host_exists,
            "zabbix.host_get": mock_host_get,
            "zabbix.hostinterface_get": mock_hostinterface_get,
            "zabbix.host_inventory_get": mock_host_inventory_get,
            "zabbix.host_inventory_set": mock_host_inventory_set,
        },
    ):
        # Blame Python 3.5 support for all this black magic
        host_present_ret = zabbix_host.present(
            host, groups, interfaces, inventory=inventory, **kwargs
        )
        host_present_changes = ast.literal_eval(
            host_present_ret["changes"]["inventory"]
        )
        assert host_present_changes == ast.literal_eval(ret["changes"]["inventory"])
        assert host_present_ret["comment"] == "Host new_host updated."
        assert host_present_ret["name"] == "new_host"
        assert host_present_ret["result"] is True
        # When Python 3.5 is gone, the following line does the job:
        # assert (
        #    zabbix_host.present(host, groups, interfaces, inventory=inventory, **kwargs)
        #    == ret
        # )
        mock_host_inventory_set.assert_called_with(
            "31337",
            _connection_password="XXXXXXXXXX",
            _connection_url="http://XXXXXXXXX/zabbix/api_jsonrpc.php",
            _connection_user="XXXXXXXXXX",
            asset_tag="ABC12345",
            clear_old=True,
            inventory_mode="0",
            vendor="TrueVendor",
        )


def test_update_inventory_keys(basic_host_configuration, existing_host_responses):
    """
    Tests the update of a inventory keys
    """
    host, groups, interfaces, kwargs, _ = basic_host_configuration
    (
        host_get_output,
        hostgroup_get_output,
        hostinterface_get_output,
        host_inventory_get_output,
    ) = existing_host_responses

    host_inventory_get_output = {
        "vendor": "FakeVendor",
        "asset_tag": "ABC12345",
    }
    hostgroup_get_output = [
        {"groupid": "16", "name": "Testing Group", "internal": "0", "flags": "0"}
    ]
    host_exists_output = True
    host_inventory_set_output = {"result": {"hostids": ["31337"]}}

    inventory = (
        {"vendor": "TrueVendor"},
        {"serialno_a": "123751236JJ123K"},
    )
    ret = {
        "changes": {
            "inventory": "{'vendor': 'TrueVendor', 'serialno_a': '123751236JJ123K'}"
        },
        "comment": "Host new_host updated.",
        "name": "new_host",
        "result": True,
    }

    mock_hostgroup_get = MagicMock(return_value=hostgroup_get_output)
    mock_host_exists = MagicMock(return_value=host_exists_output)
    mock_host_get = MagicMock(return_value=host_get_output)
    mock_hostinterface_get = MagicMock(return_value=hostinterface_get_output)
    mock_host_inventory_get = MagicMock(return_value=host_inventory_get_output)
    mock_host_inventory_set = MagicMock(return_value=host_inventory_set_output)
    with patch.dict(
        zabbix_host.__salt__,
        {
            "zabbix.hostgroup_get": mock_hostgroup_get,
            "zabbix.host_exists": mock_host_exists,
            "zabbix.host_get": mock_host_get,
            "zabbix.hostinterface_get": mock_hostinterface_get,
            "zabbix.host_inventory_get": mock_host_inventory_get,
            "zabbix.host_inventory_set": mock_host_inventory_set,
        },
    ):
        # Blame Python 3.5 support for all this black magic
        host_present_ret = zabbix_host.present(
            host, groups, interfaces, inventory=inventory, **kwargs
        )
        host_present_changes = ast.literal_eval(
            host_present_ret["changes"]["inventory"]
        )
        assert host_present_changes == ast.literal_eval(ret["changes"]["inventory"])
        assert host_present_ret["comment"] == "Host new_host updated."
        assert host_present_ret["name"] == "new_host"
        assert host_present_ret["result"] is True
        # When Python 3.5 is gone, the following line does the job:
        # assert (
        #    zabbix_host.present(host, groups, interfaces, inventory=inventory, **kwargs)
        #    == ret
        # )
        mock_host_inventory_set.assert_called_with(
            "31337",
            _connection_password="XXXXXXXXXX",
            _connection_url="http://XXXXXXXXX/zabbix/api_jsonrpc.php",
            _connection_user="XXXXXXXXXX",
            clear_old=True,
            inventory_mode="0",
            serialno_a="123751236JJ123K",
            vendor="TrueVendor",
        )


def test_update_inventory_values_without_clear_existing_data(
    basic_host_configuration, existing_host_responses
):
    """
    Tests the update of an inventory value without clear the current inventory
    """
    host, groups, interfaces, kwargs, _ = basic_host_configuration
    (
        host_get_output,
        hostgroup_get_output,
        hostinterface_get_output,
        host_inventory_get_output,
    ) = existing_host_responses

    host_inventory_get_output = {
        "vendor": "FakeVendor",
        "asset_tag": "ABC12345",
    }
    hostgroup_get_output = [
        {"groupid": "16", "name": "Testing Group", "internal": "0", "flags": "0"}
    ]
    host_exists_output = True
    host_inventory_set_output = {"result": {"hostids": ["31337"]}}

    inventory = ({"vendor": "TrueVendor"},)
    ret = {
        "changes": {"inventory": "{'vendor': 'TrueVendor'}"},
        "comment": "Host new_host updated.",
        "name": "new_host",
        "result": True,
    }

    mock_hostgroup_get = MagicMock(return_value=hostgroup_get_output)
    mock_host_exists = MagicMock(return_value=host_exists_output)
    mock_host_get = MagicMock(return_value=host_get_output)
    mock_hostinterface_get = MagicMock(return_value=hostinterface_get_output)
    mock_host_inventory_get = MagicMock(return_value=host_inventory_get_output)
    mock_host_inventory_set = MagicMock(return_value=host_inventory_set_output)
    with patch.dict(
        zabbix_host.__salt__,
        {
            "zabbix.hostgroup_get": mock_hostgroup_get,
            "zabbix.host_exists": mock_host_exists,
            "zabbix.host_get": mock_host_get,
            "zabbix.hostinterface_get": mock_hostinterface_get,
            "zabbix.host_inventory_get": mock_host_inventory_get,
            "zabbix.host_inventory_set": mock_host_inventory_set,
        },
    ):
        # Blame Python 3.5 support for all this black magic
        host_present_ret = zabbix_host.present(
            host,
            groups,
            interfaces,
            inventory=inventory,
            inventory_clean=False,
            **kwargs
        )
        host_present_changes = ast.literal_eval(
            host_present_ret["changes"]["inventory"]
        )
        assert host_present_changes == ast.literal_eval(ret["changes"]["inventory"])
        assert host_present_ret["comment"] == "Host new_host updated."
        assert host_present_ret["name"] == "new_host"
        assert host_present_ret["result"] is True
        # When Python 3.5 is gone, the following line does the job:
        # assert (
        #    zabbix_host.present(
        #        host,
        #        groups,
        #        interfaces,
        #        inventory=inventory,
        #        inventory_clean=False,
        #        **kwargs
        #    )
        #    == ret
        # )
        mock_host_inventory_set.assert_called_with(
            "31337",
            _connection_password="XXXXXXXXXX",
            _connection_url="http://XXXXXXXXX/zabbix/api_jsonrpc.php",
            _connection_user="XXXXXXXXXX",
            clear_old=False,
            inventory_mode="0",
            vendor="TrueVendor",
        )


def test_ensure_nothing_happens_when_inventory_is_not_sent(
    basic_host_configuration, existing_host_responses
):
    """
    Test to ensure that the current inventory is not erased when inventory is not sent
    """
    host, groups, interfaces, kwargs, _ = basic_host_configuration
    (
        host_get_output,
        hostgroup_get_output_up,
        hostinterface_get_output,
        _,
    ) = existing_host_responses

    host_inventory_get_output = {
        "vendor": "FakeVendor",
        "asset_tag": "ABC12345",
    }
    hostgroup_get_output = [
        [{"groupid": "16", "name": "Testing Group", "internal": "0", "flags": "0"}],
        hostgroup_get_output_up,
    ]
    host_exists_output = True

    ret = {
        "changes": {},
        "comment": "Host new_host already exists.",
        "name": "new_host",
        "result": True,
    }

    mock_hostgroup_get = MagicMock(side_effect=hostgroup_get_output)
    mock_host_exists = MagicMock(return_value=host_exists_output)
    mock_host_get = MagicMock(return_value=host_get_output)
    mock_hostinterface_get = MagicMock(return_value=hostinterface_get_output)
    mock_host_inventory_get = MagicMock(return_value=host_inventory_get_output)
    mock_host_inventory_set = MagicMock(return_value=False)
    with patch.dict(
        zabbix_host.__salt__,
        {
            "zabbix.hostgroup_get": mock_hostgroup_get,
            "zabbix.host_exists": mock_host_exists,
            "zabbix.host_get": mock_host_get,
            "zabbix.hostinterface_get": mock_hostinterface_get,
            "zabbix.host_inventory_get": mock_host_inventory_get,
            "zabbix.host_inventory_set": mock_host_inventory_set,
        },
    ):
        assert zabbix_host.present(host, groups, interfaces, **kwargs) == ret
        assert (
            not mock_host_inventory_set.called
        ), "host_inventory_set should not be called"


def test_ensure_that_inventory_is_not_sent_when_inventory_disabled(
    basic_host_configuration, existing_host_responses
):
    """
    Test to ensure that the inventory is not set when inventory_mode is disabled
    """
    host, groups, interfaces, kwargs, _ = basic_host_configuration
    (
        host_get_output,
        hostgroup_get_output_up,
        hostinterface_get_output,
        host_inventory_get_output,
    ) = existing_host_responses

    hostgroup_get_output = [
        [{"groupid": "16", "name": "Testing Group", "internal": "0", "flags": "0"}],
        hostgroup_get_output_up,
    ]
    host_exists_output = True

    kwargs["inventory_mode"] = "-1"
    inventory = ({"vendor": "TrueVendor"},)
    ret = {
        "changes": {},
        "comment": "Host new_host already exists.",
        "name": "new_host",
        "result": True,
    }

    mock_hostgroup_get = MagicMock(side_effect=hostgroup_get_output)
    mock_host_exists = MagicMock(return_value=host_exists_output)
    mock_host_get = MagicMock(return_value=host_get_output)
    mock_hostinterface_get = MagicMock(return_value=hostinterface_get_output)
    mock_host_inventory_get = MagicMock(return_value=host_inventory_get_output)
    mock_host_inventory_set = MagicMock(return_value=False)
    with patch.dict(
        zabbix_host.__salt__,
        {
            "zabbix.hostgroup_get": mock_hostgroup_get,
            "zabbix.host_exists": mock_host_exists,
            "zabbix.host_get": mock_host_get,
            "zabbix.hostinterface_get": mock_hostinterface_get,
            "zabbix.host_inventory_get": mock_host_inventory_get,
            "zabbix.host_inventory_set": mock_host_inventory_set,
        },
    ):
        assert (
            zabbix_host.present(host, groups, interfaces, inventory=inventory, **kwargs)
            == ret
        )
        assert (
            not mock_host_inventory_set.called
        ), "host_inventory_set should not be called"


def test_update_inventory_and_restore_inventory_mode(
    basic_host_configuration, existing_host_responses
):
    """
    Tests the restore of inventory_mode to automatic after update
    """
    host, groups, interfaces, kwargs, _ = basic_host_configuration
    (
        host_get_output,
        hostgroup_get_output,
        hostinterface_get_output,
        host_inventory_get_output,
    ) = existing_host_responses

    host_get_output[0]["inventory_mode"] = "1"
    host_inventory_get_output = {
        "vendor": "FakeVendor",
        "asset_tag": "ABC12345",
    }
    hostgroup_get_output = [
        {"groupid": "16", "name": "Testing Group", "internal": "0", "flags": "0"}
    ]
    host_exists_output = True
    host_inventory_set_output = {"result": {"hostids": ["31337"]}}
    host_update_output = "31337"

    kwargs["inventory_mode"] = "1"
    inventory = (
        {"vendor": "TrueVendor"},
        {"asset_tag": "ABC12345"},
    )

    ret = {
        "changes": {"inventory": "{'vendor': 'TrueVendor', 'asset_tag': 'ABC12345'}"},
        "comment": "Host new_host updated.",
        "name": "new_host",
        "result": True,
    }

    mock_hostgroup_get = MagicMock(return_value=hostgroup_get_output)
    mock_host_exists = MagicMock(return_value=host_exists_output)
    mock_host_get = MagicMock(return_value=host_get_output)
    mock_hostinterface_get = MagicMock(return_value=hostinterface_get_output)
    mock_host_inventory_get = MagicMock(return_value=host_inventory_get_output)
    mock_host_inventory_set = MagicMock(return_value=host_inventory_set_output)
    with patch.dict(
        zabbix_host.__salt__,
        {
            "zabbix.hostgroup_get": mock_hostgroup_get,
            "zabbix.host_exists": mock_host_exists,
            "zabbix.host_get": mock_host_get,
            "zabbix.hostinterface_get": mock_hostinterface_get,
            "zabbix.host_inventory_get": mock_host_inventory_get,
            "zabbix.host_inventory_set": mock_host_inventory_set,
        },
    ):
        # Blame Python 3.5 support for all this black magic
        host_present_ret = zabbix_host.present(
            host, groups, interfaces, inventory=inventory, **kwargs
        )
        host_present_changes = ast.literal_eval(
            host_present_ret["changes"]["inventory"]
        )
        assert host_present_changes == ast.literal_eval(ret["changes"]["inventory"])
        assert host_present_ret["comment"] == "Host new_host updated."
        assert host_present_ret["name"] == "new_host"
        assert host_present_ret["result"] is True
        # When Python 3.5 is gone, the following line does the job:
        # assert (
        #    zabbix_host.present(host, groups, interfaces, inventory=inventory, **kwargs)
        #    == ret
        # )
        mock_host_inventory_set.assert_called_with(
            "31337",
            _connection_password="XXXXXXXXXX",
            _connection_url="http://XXXXXXXXX/zabbix/api_jsonrpc.php",
            _connection_user="XXXXXXXXXX",
            asset_tag="ABC12345",
            clear_old=True,
            inventory_mode="1",
            vendor="TrueVendor",
        )


def test_clear_inventory_value_sending_an_empty_key(
    basic_host_configuration, existing_host_responses
):
    """
    Tests clear an specific value sending it as an empty string
    """
    host, groups, interfaces, kwargs, _ = basic_host_configuration
    (
        host_get_output,
        hostgroup_get_output,
        hostinterface_get_output,
        host_inventory_get_output,
    ) = existing_host_responses

    host_inventory_get_output = {
        "vendor": "FakeVendor",
        "asset_tag": "ABC12345",
    }
    hostgroup_get_output = [
        {"groupid": "16", "name": "Testing Group", "internal": "0", "flags": "0"}
    ]
    host_exists_output = True
    host_inventory_set_output = {"result": {"hostids": ["31337"]}}

    inventory = (
        {"vendor": "TrueVendor"},
        {"asset_tag": ""},
    )
    ret = {
        "changes": {"inventory": "{'vendor': 'TrueVendor', 'asset_tag': ''}"},
        "comment": "Host new_host updated.",
        "name": "new_host",
        "result": True,
    }

    mock_hostgroup_get = MagicMock(return_value=hostgroup_get_output)
    mock_host_exists = MagicMock(return_value=host_exists_output)
    mock_host_get = MagicMock(return_value=host_get_output)
    mock_hostinterface_get = MagicMock(return_value=hostinterface_get_output)
    mock_host_inventory_get = MagicMock(return_value=host_inventory_get_output)
    mock_host_inventory_set = MagicMock(return_value=host_inventory_set_output)
    with patch.dict(
        zabbix_host.__salt__,
        {
            "zabbix.hostgroup_get": mock_hostgroup_get,
            "zabbix.host_exists": mock_host_exists,
            "zabbix.host_get": mock_host_get,
            "zabbix.hostinterface_get": mock_hostinterface_get,
            "zabbix.host_inventory_get": mock_host_inventory_get,
            "zabbix.host_inventory_set": mock_host_inventory_set,
        },
    ):
        # Blame Python 3.5 support for all this black magic
        host_present_ret = zabbix_host.present(
            host,
            groups,
            interfaces,
            inventory=inventory,
            inventory_clean=False,
            **kwargs
        )
        host_present_changes = ast.literal_eval(
            host_present_ret["changes"]["inventory"]
        )
        assert host_present_changes == ast.literal_eval(ret["changes"]["inventory"])
        assert host_present_ret["comment"] == "Host new_host updated."
        assert host_present_ret["name"] == "new_host"
        assert host_present_ret["result"] is True
        # When Python 3.5 is gone, the following line does the job:
        # assert (
        #    zabbix_host.present(
        #        host,
        #        groups,
        #        interfaces,
        #        inventory=inventory,
        #        inventory_clean=False,
        #        **kwargs
        #    )
        #    == ret
        # )
        mock_host_inventory_set.assert_called_with(
            "31337",
            _connection_password="XXXXXXXXXX",
            _connection_url="http://XXXXXXXXX/zabbix/api_jsonrpc.php",
            _connection_user="XXXXXXXXXX",
            asset_tag="",
            clear_old=False,
            inventory_mode="0",
            vendor="TrueVendor",
        )



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import io

import salt.modules.hosts as hosts
import salt.utils.data
import salt.utils.platform
import salt.utils.stringutils
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, mock_open, patch
from tests.support.unit import TestCase


class HostsTestCase(TestCase, LoaderModuleMockMixin):
    """
    TestCase for salt.modules.hosts
    """

    def setup_loader_modules(self):
        return {hosts: {}}

    # 'list_hosts' function tests: 1

    def test_list_hosts(self):
        """
        Tests return the hosts found in the hosts file
        """
        with patch(
            "salt.modules.hosts._list_hosts",
            MagicMock(return_value={"10.10.10.10": {"aliases": ["Salt1", "Salt2"]}}),
        ):
            self.assertDictEqual(
                {"10.10.10.10": {"aliases": ["Salt1", "Salt2"]}}, hosts.list_hosts()
            )

    # 'get_ip' function tests: 3

    def test_get_ip(self):
        """
        Tests return ip associated with the named host
        """
        with patch(
            "salt.modules.hosts._list_hosts",
            MagicMock(return_value={"10.10.10.10": {"aliases": ["Salt1", "Salt2"]}}),
        ):
            self.assertEqual("10.10.10.10", hosts.get_ip("Salt1"))

            self.assertEqual("", hosts.get_ip("Salt3"))

    def test_get_ip_none(self):
        """
        Tests return ip associated with the named host
        """
        with patch("salt.modules.hosts._list_hosts", MagicMock(return_value="")):
            self.assertEqual("", hosts.get_ip("Salt1"))

    # 'get_alias' function tests: 2

    def test_get_alias(self):
        """
        Tests return the list of aliases associated with an ip
        """
        with patch(
            "salt.modules.hosts._list_hosts",
            MagicMock(return_value={"10.10.10.10": {"aliases": ["Salt1", "Salt2"]}}),
        ):
            self.assertListEqual(["Salt1", "Salt2"], hosts.get_alias("10.10.10.10"))

    def test_get_alias_none(self):
        """
        Tests return the list of aliases associated with an ip
        """
        with patch(
            "salt.modules.hosts._list_hosts",
            MagicMock(return_value={"10.10.10.10": {"aliases": ["Salt1", "Salt2"]}}),
        ):
            self.assertListEqual([], hosts.get_alias("10.10.10.11"))

    # 'has_pair' function tests: 1

    def test_has_pair(self):
        """
        Tests return True / False if the alias is set
        """
        with patch(
            "salt.modules.hosts._list_hosts",
            MagicMock(return_value={"10.10.10.10": {"aliases": ["Salt1", "Salt2"]}}),
        ):
            self.assertTrue(hosts.has_pair("10.10.10.10", "Salt1"))

            self.assertFalse(hosts.has_pair("10.10.10.10", "Salt3"))

    # 'set_host' function tests: 3

    def test_set_host(self):
        """
        Tests true if the alias is set
        """
        hosts_file = "/etc/hosts"
        if salt.utils.platform.is_windows():
            hosts_file = r"C:\Windows\System32\Drivers\etc\hosts"

        with patch(
            "salt.modules.hosts.__get_hosts_filename",
            MagicMock(return_value=hosts_file),
        ), patch("os.path.isfile", MagicMock(return_value=False)), patch.dict(
            hosts.__salt__, {"config.option": MagicMock(return_value=None)}
        ):
            self.assertFalse(hosts.set_host("10.10.10.10", "Salt1"))

    def test_set_host_true(self):
        """
        Tests true if the alias is set
        """
        with patch(
            "salt.modules.hosts.__get_hosts_filename",
            MagicMock(return_value="/etc/hosts"),
        ), patch("os.path.isfile", MagicMock(return_value=True)), patch(
            "salt.utils.files.fopen", mock_open(b"")
        ):
            mock_opt = MagicMock(return_value=None)
            with patch.dict(hosts.__salt__, {"config.option": mock_opt}):
                self.assertTrue(hosts.set_host("10.10.10.10", "Salt1"))

    def test_set_host_true_remove(self):
        """
        Test if an empty hosts value removes existing entries
        """
        with patch(
            "salt.modules.hosts.__get_hosts_filename",
            MagicMock(return_value="/etc/hosts"),
        ), patch("os.path.isfile", MagicMock(return_value=True)):
            data = [
                "\n".join(
                    (
                        "1.1.1.1 foo.foofoo foo",
                        "2.2.2.2 bar.barbar bar",
                        "3.3.3.3 asdf.asdfadsf asdf",
                        "1.1.1.1 foofoo.foofoo foofoo",
                    )
                )
            ]

            class TmpStringIO(io.StringIO):
                def __init__(self, fn, mode="r"):
                    self.mode = mode
                    initial_value = data[0]
                    if "w" in self.mode:
                        initial_value = ""
                    super().__init__(initial_value)

                def __enter__(self):
                    return self

                def __exit__(self, exc_type, exc_value, traceback):
                    self.close()

                def close(self):
                    # Don't save unless there's something there. In Windows
                    # the class gets initialized the first time with mode = w
                    # which sets the initial value to ''. When the class closes
                    # it clears out data and causes the test to fail.
                    # I don't know why it get's initialized with a mode of 'w'
                    # For the purposes of this test data shouldn't be empty
                    # This is a problem with this class and not with the hosts
                    # module
                    if self.getvalue():
                        data[0] = self.getvalue()
                    io.StringIO.close(self)

                def read(self, *args):
                    ret = super().read(*args)
                    if "b" in self.mode:
                        return salt.utils.stringutils.to_bytes(ret)
                    else:
                        return ret

                def write(self, s, *args):
                    if "b" in self.mode:
                        if not isinstance(s, bytes):
                            # Make this act like a binary filehandle
                            raise TypeError(
                                "a bytes-like object is required, not 'str'"
                            )
                        # The StringIO wants a str type, it won't take
                        # bytes. Convert before writing to it.
                        return super().write(salt.utils.stringutils.to_str(s), *args)
                    else:
                        if not isinstance(s, str):
                            # Make this act like a non-binary filehandle
                            raise TypeError("write() argument must be str, not bytes")
                    return super().write(s, *args)

                def readlines(self):
                    ret = super().readlines()
                    if "b" in self.mode:
                        return salt.utils.data.encode(ret)
                    else:
                        return ret

                def writelines(self, lines):
                    for line in lines:
                        self.write(line)

            expected = (
                "\n".join(
                    (
                        "2.2.2.2 bar.barbar bar",
                        "3.3.3.3 asdf.asdfadsf asdf",
                    )
                )
                + "\n"
            )

            with patch("salt.utils.files.fopen", TmpStringIO):
                mock_opt = MagicMock(return_value=None)
                with patch.dict(hosts.__salt__, {"config.option": mock_opt}):
                    self.assertTrue(hosts.set_host("1.1.1.1", " "))

            self.assertEqual(data[0], expected)

    # 'rm_host' function tests: 2

    def test_rm_host(self):
        """
        Tests if specified host entry gets removed from the hosts file
        """
        hosts_content = (
            b"# one line comment\n10.10.10.10    Salt1\n9.9.9.9    Salt2   # comment\n"
        )
        with patch("salt.utils.files.fopen", mock_open(hosts_content)), patch(
            "salt.modules.hosts.__get_hosts_filename",
            MagicMock(return_value="/etc/hosts"),
        ), patch("salt.modules.hosts.has_pair", MagicMock(return_value=True)), patch(
            "os.path.isfile", MagicMock(return_value=True)
        ):
            mock_opt = MagicMock(return_value=None)
            with patch.dict(hosts.__salt__, {"config.option": mock_opt}):
                self.assertTrue(hosts.rm_host("10.10.10.10", "Salt1"))

    def test_rm_host_false(self):
        """
        Tests if specified host entry gets removed from the hosts file
        """
        with patch("salt.modules.hosts.has_pair", MagicMock(return_value=False)):
            self.assertTrue(hosts.rm_host("10.10.10.10", "Salt1"))

    # 'add_host' function tests: 3

    def test_add_host(self):
        """
        Tests if specified host entry gets added from the hosts file
        """
        hosts_file = "/etc/hosts"
        if salt.utils.platform.is_windows():
            hosts_file = r"C:\Windows\System32\Drivers\etc\hosts"

        with patch("salt.utils.files.fopen", mock_open()), patch(
            "salt.modules.hosts.__get_hosts_filename",
            MagicMock(return_value=hosts_file),
        ):
            mock_opt = MagicMock(return_value=None)
            with patch.dict(hosts.__salt__, {"config.option": mock_opt}):
                self.assertTrue(hosts.add_host("10.10.10.10", "Salt1"))

    def test_add_host_no_file(self):
        """
        Tests if specified host entry gets added from the hosts file
        """
        with patch("salt.utils.files.fopen", mock_open()), patch(
            "os.path.isfile", MagicMock(return_value=False)
        ):
            mock_opt = MagicMock(return_value=None)
            with patch.dict(hosts.__salt__, {"config.option": mock_opt}):
                self.assertFalse(hosts.add_host("10.10.10.10", "Salt1"))

    def test_add_host_create_entry(self):
        """
        Tests if specified host entry gets added from the hosts file
        """
        with patch("salt.utils.files.fopen", mock_open()), patch(
            "os.path.isfile", MagicMock(return_value=True)
        ):
            mock_opt = MagicMock(return_value=None)
            with patch.dict(hosts.__salt__, {"config.option": mock_opt}):
                self.assertTrue(hosts.add_host("10.10.10.10", "Salt1"))

    def test_set_comment(self):
        """
        Tests return True / False when setting a comment
        """
        hosts_file = "/etc/hosts"
        if salt.utils.platform.is_windows():
            hosts_file = r"C:\Windows\System32\Drivers\etc\hosts"

        with patch("salt.utils.files.fopen", mock_open()), patch(
            "salt.modules.hosts.__get_hosts_filename",
            MagicMock(return_value=hosts_file),
        ):
            mock_opt = MagicMock(return_value=None)
            with patch.dict(hosts.__salt__, {"config.option": mock_opt}):
                with patch(
                    "salt.modules.hosts._list_hosts",
                    MagicMock(
                        return_value={"10.10.10.10": {"aliases": ["Salt1", "Salt2"]}}
                    ),
                ):
                    self.assertTrue(hosts.set_comment("10.10.10.10", "A comment"))

                    self.assertFalse(hosts.set_comment("10.10.10.11", "A comment"))



import logging
import os
import signal

import pytest

import salt.utils.vt as vt


@pytest.mark.skip_on_windows(reason="salt.utils.vt.Terminal doesn't have _spawn.")
def test_isalive_no_child():
    term = vt.Terminal(
        "sleep 100",
        shell=True,
        stream_stdout=False,
        stream_stderr=False,
    )

    # make sure we have a valid term before we kill the term
    # commenting out for now, terminal seems to be stopping before this point
    aliveness = term.isalive()
    assert term.exitstatus is None
    assert aliveness is True
    # use a large hammer to make sure pid is really dead which will cause it to
    # raise an exception that we want to test for.
    os.kill(term.pid, signal.SIGKILL)
    os.waitpid(term.pid, 0)
    aliveness = term.isalive()
    assert term.exitstatus == 0
    assert aliveness is False


@pytest.mark.parametrize("test_cmd", ["echo", "ls"])
@pytest.mark.skip_on_windows()
def test_log_sanitize(test_cmd, caplog):
    """
    test when log_sanitize is passed in
    we do not see the password in either
    standard out or standard error logs
    """
    password = "123456"
    cmd = [test_cmd, password]
    term = vt.Terminal(
        cmd,
        log_stdout=True,
        log_stderr=True,
        log_stdout_level="debug",
        log_stderr_level="debug",
        log_sanitize=password,
        stream_stdout=False,
        stream_stderr=False,
    )
    with caplog.at_level(logging.DEBUG):
        term.recv()
    assert password not in caplog.text
    assert "******" in caplog.text



"""
Unit tests for the Highstate Returner Cache.
"""

import io
import json

import pytest

import salt.returners.highstate_return as highstate
import salt.utils.files


@pytest.fixture
def output_file(tmp_path):
    return tmp_path / "output-file"


@pytest.fixture
def configure_loader_modules(output_file):
    return {
        highstate: {
            "__opts__": {
                "highstate.report_everything": True,
                "highstate.report_format": "json",
                "highstate.report_delivery": "file",
                "highstate.file_output": str(output_file),
            }
        }
    }


def test_generate_table_should_correctly_escape_html_characters_when_data_contains_both_list_and_dict():
    unescaped_fnord = "&fnord&"
    unescaped_dronf = "<dronf>"
    expected_escaped_fnord = "&amp;fnord&amp;"
    expected_escaped_dronf = "&lt;dronf&gt;"
    data = [["something", "or", "another", unescaped_fnord, {"cool": unescaped_dronf}]]

    out = io.StringIO()
    highstate._generate_html_table(data=data, out=out)
    out.seek(0)
    actual_table = out.read()

    assert expected_escaped_fnord in actual_table
    assert expected_escaped_dronf in actual_table


def test_pipe_in_name(output_file):
    ret = {
        "fun_args": ["test"],
        "jid": "20180308201402941603",
        "return": {
            "cmd_|-test_|-echo hi | grep h\n_|-run": {
                "comment": 'Command "echo hi | grep h\n" run',
                "name": "echo hi | grep h\n",
                "start_time": "20:14:03.053612",
                "result": True,
                "duration": 75.198,
                "__run_num__": 0,
                "__sls__": "test",
                "changes": {
                    "pid": 1429,
                    "retcode": 0,
                    "stderr": "",
                    "stdout": "hi",
                },
                "__id__": "test",
            }
        },
        "retcode": 0,
        "success": True,
        "fun": "state.apply",
        "id": "salt",
        "out": "highstate",
    }
    expected = [
        {
            "stats": [
                {"total": 1},
                {"failed": 0, "__style__": "failed"},
                {"unchanged": 0, "__style__": "unchanged"},
                {"changed": 1, "__style__": "changed"},
                {"duration": 75.198},
            ],
        },
        {
            "job": [
                {"function": "state.apply"},
                {"arguments": ["test"]},
                {"jid": "20180308201402941603"},
                {"success": True},
                {"retcode": 0},
            ],
        },
        {
            "states": [
                {
                    "test": [
                        {"function": "cmd.run"},
                        {"name": "echo hi | grep h\n"},
                        {"result": True},
                        {"duration": 75.198},
                        {"comment": 'Command "echo hi | grep h\n" run'},
                        {
                            "changes": [
                                {"pid": 1429},
                                {"retcode": 0},
                                {"stderr": ""},
                                {"stdout": "hi"},
                            ]
                        },
                        {"started": "20:14:03.053612"},
                    ],
                    "__style__": "changed",
                }
            ]
        },
    ]
    highstate.returner(ret)
    with salt.utils.files.fopen(str(output_file)) as fh_:
        assert json.load(fh_) == expected



"""
Tests for salt.utils.dateutils
"""

import datetime

import pytest

import salt.utils.dateutils
from tests.support.mock import patch
from tests.support.unit import TestCase

try:
    import timelib  # pylint: disable=import-error,unused-import

    HAS_TIMELIB = True
except ImportError:
    HAS_TIMELIB = False


class DateutilsTestCase(TestCase):
    def test_date_cast(self):
        now = datetime.datetime.now()
        with patch("datetime.datetime"):
            datetime.datetime.now.return_value = now
            self.assertEqual(now, salt.utils.dateutils.date_cast(None))
        self.assertEqual(now, salt.utils.dateutils.date_cast(now))
        try:
            ret = salt.utils.dateutils.date_cast("Mon Dec 23 10:19:15 MST 2013")
            expected_ret = datetime.datetime(2013, 12, 23, 10, 19, 15)
            self.assertEqual(ret, expected_ret)
        except RuntimeError:
            if not HAS_TIMELIB:
                # Unparseable without timelib installed
                self.skipTest("'timelib' is not installed")
            else:
                raise

    @pytest.mark.skipif(not HAS_TIMELIB, reason="'timelib' is not installed")
    def test_strftime(self):

        # Taken from doctests

        expected_ret = "2002-12-25"

        src = datetime.datetime(2002, 12, 25, 12, 00, 00, 00)
        ret = salt.utils.dateutils.strftime(src)
        self.assertEqual(ret, expected_ret)

        src = "2002/12/25"
        ret = salt.utils.dateutils.strftime(src)
        self.assertEqual(ret, expected_ret)

        src = 1040814000
        ret = salt.utils.dateutils.strftime(src)
        self.assertEqual(ret, expected_ret)

        src = "1040814000"
        ret = salt.utils.dateutils.strftime(src)
        self.assertEqual(ret, expected_ret)



"""
    :codeauthor: Rahul Handay <rahulha@saltstack.com>
"""

import os

import pytest

import salt.modules.systemd_service as systemd
import salt.utils.systemd
from salt.exceptions import CommandExecutionError
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase

_SYSTEMCTL_STATUS = {
    "sshd.service": {
        "stdout": """\
* sshd.service - OpenSSH Daemon
   Loaded: loaded (/usr/lib/systemd/system/sshd.service; disabled; vendor preset: disabled)
   Active: inactive (dead)""",
        "stderr": "",
        "retcode": 3,
        "pid": 12345,
    },
    "foo.service": {
        "stdout": """\
* foo.service
   Loaded: not-found (Reason: No such file or directory)
   Active: inactive (dead)""",
        "stderr": "",
        "retcode": 3,
        "pid": 12345,
    },
}

# This reflects systemd >= 231 behavior
_SYSTEMCTL_STATUS_GTE_231 = {
    "bar.service": {
        "stdout": "Unit bar.service could not be found.",
        "stderr": "",
        "retcode": 4,
        "pid": 12345,
    },
}

_LIST_UNIT_FILES = """\
service1.service                           enabled              -
service2.service                           disabled             -
service3.service                           static               -
timer1.timer                               enabled              -
timer2.timer                               disabled             -
timer3.timer                               static               -
service4.service                           enabled              enabled
service5.service                           disabled             enabled
service6.service                           static               enabled
timer4.timer                               enabled              enabled
timer5.timer                               disabled             enabled
timer6.timer                               static               enabled
service7.service                           enabled              disabled
service8.service                           disabled             disabled
service9.service                           static               disabled
timer7.timer                               enabled              disabled
timer8.timer                               disabled             disabled
timer9.timer                               static               disabled
service10.service                          enabled
service11.service                          disabled
service12.service                          static
timer10.timer                              enabled
timer11.timer                              disabled
timer12.timer                              static"""


class SystemdTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test case for salt.modules.systemd
    """

    def setup_loader_modules(self):
        return {systemd: {}}

    def test_systemctl_reload(self):
        """
        Test to Reloads systemctl
        """
        mock = MagicMock(
            side_effect=[
                {"stdout": "Who knows why?", "stderr": "", "retcode": 1, "pid": 12345},
                {"stdout": "", "stderr": "", "retcode": 0, "pid": 54321},
            ]
        )
        with patch.dict(systemd.__salt__, {"cmd.run_all": mock}):
            self.assertRaisesRegex(
                CommandExecutionError,
                "Problem performing systemctl daemon-reload: Who knows why?",
                systemd.systemctl_reload,
            )
            self.assertTrue(systemd.systemctl_reload())

    def test_get_enabled(self):
        """
        Test to return a list of all enabled services
        """
        cmd_mock = MagicMock(return_value=_LIST_UNIT_FILES)
        listdir_mock = MagicMock(return_value=["foo", "bar", "baz", "README"])
        sd_mock = MagicMock(
            return_value={x.replace(".service", "") for x in _SYSTEMCTL_STATUS}
        )
        access_mock = MagicMock(
            side_effect=lambda x, y: x
            != os.path.join(systemd.INITSCRIPT_PATH, "README")
        )
        sysv_enabled_mock = MagicMock(side_effect=lambda x, _: x == "baz")

        with patch.dict(systemd.__salt__, {"cmd.run": cmd_mock}):
            with patch.object(os, "listdir", listdir_mock):
                with patch.object(systemd, "_get_systemd_services", sd_mock):
                    with patch.object(os, "access", side_effect=access_mock):
                        with patch.object(systemd, "_sysv_enabled", sysv_enabled_mock):
                            self.assertListEqual(
                                systemd.get_enabled(),
                                [
                                    "baz",
                                    "service1",
                                    "service10",
                                    "service4",
                                    "service7",
                                    "timer1.timer",
                                    "timer10.timer",
                                    "timer4.timer",
                                    "timer7.timer",
                                ],
                            )

    def test_get_disabled(self):
        """
        Test to return a list of all disabled services
        """
        cmd_mock = MagicMock(return_value=_LIST_UNIT_FILES)
        # 'foo' should collide with the systemd services (as returned by
        # sd_mock) and thus not be returned by _get_sysv_services(). It doesn't
        # matter that it's not part of the _LIST_UNIT_FILES output, we just
        # want to ensure that 'foo' isn't identified as a disabled initscript
        # even though below we are mocking it to show as not enabled (since
        # only 'baz' will be considered an enabled sysv service).
        listdir_mock = MagicMock(return_value=["foo", "bar", "baz", "README"])
        sd_mock = MagicMock(
            return_value={x.replace(".service", "") for x in _SYSTEMCTL_STATUS}
        )
        access_mock = MagicMock(
            side_effect=lambda x, y: x
            != os.path.join(systemd.INITSCRIPT_PATH, "README")
        )
        sysv_enabled_mock = MagicMock(side_effect=lambda x, _: x == "baz")

        with patch.dict(systemd.__salt__, {"cmd.run": cmd_mock}):
            with patch.object(os, "listdir", listdir_mock):
                with patch.object(systemd, "_get_systemd_services", sd_mock):
                    with patch.object(os, "access", side_effect=access_mock):
                        with patch.object(systemd, "_sysv_enabled", sysv_enabled_mock):
                            self.assertListEqual(
                                systemd.get_disabled(),
                                [
                                    "bar",
                                    "service11",
                                    "service2",
                                    "service5",
                                    "service8",
                                    "timer11.timer",
                                    "timer2.timer",
                                    "timer5.timer",
                                    "timer8.timer",
                                ],
                            )

    def test_get_static(self):
        """
        Test to return a list of all disabled services
        """
        cmd_mock = MagicMock(return_value=_LIST_UNIT_FILES)
        # 'foo' should collide with the systemd services (as returned by
        # sd_mock) and thus not be returned by _get_sysv_services(). It doesn't
        # matter that it's not part of the _LIST_UNIT_FILES output, we just
        # want to ensure that 'foo' isn't identified as a disabled initscript
        # even though below we are mocking it to show as not enabled (since
        # only 'baz' will be considered an enabled sysv service).
        listdir_mock = MagicMock(return_value=["foo", "bar", "baz", "README"])
        sd_mock = MagicMock(
            return_value={x.replace(".service", "") for x in _SYSTEMCTL_STATUS}
        )
        access_mock = MagicMock(
            side_effect=lambda x, y: x
            != os.path.join(systemd.INITSCRIPT_PATH, "README")
        )
        sysv_enabled_mock = MagicMock(side_effect=lambda x, _: x == "baz")

        with patch.dict(systemd.__salt__, {"cmd.run": cmd_mock}):
            with patch.object(os, "listdir", listdir_mock):
                with patch.object(systemd, "_get_systemd_services", sd_mock):
                    with patch.object(os, "access", side_effect=access_mock):
                        with patch.object(systemd, "_sysv_enabled", sysv_enabled_mock):
                            self.assertListEqual(
                                systemd.get_static(),
                                [
                                    "service12",
                                    "service3",
                                    "service6",
                                    "service9",
                                    "timer12.timer",
                                    "timer3.timer",
                                    "timer6.timer",
                                    "timer9.timer",
                                ],
                            )

    def test_get_all(self):
        """
        Test to return a list of all available services
        """
        listdir_mock = MagicMock(
            side_effect=[
                ["foo.service", "multi-user.target.wants", "mytimer.timer"],
                [],
                ["foo.service", "multi-user.target.wants", "bar.service"],
                ["mysql", "nginx", "README"],
            ]
        )
        access_mock = MagicMock(
            side_effect=lambda x, y: x
            != os.path.join(systemd.INITSCRIPT_PATH, "README")
        )
        with patch.object(os, "listdir", listdir_mock):
            with patch.object(os, "access", side_effect=access_mock):
                self.assertListEqual(
                    systemd.get_all(), ["bar", "foo", "mysql", "mytimer.timer", "nginx"]
                )

    def test_available(self):
        """
        Test to check that the given service is available
        """
        mock = MagicMock(side_effect=lambda x: _SYSTEMCTL_STATUS[x])

        # systemd < 231
        with patch.dict(systemd.__context__, {"salt.utils.systemd.version": 230}):
            with patch.object(systemd, "_systemctl_status", mock), patch.object(
                systemd, "offline", MagicMock(return_value=False)
            ):
                self.assertTrue(systemd.available("sshd.service"))
                self.assertFalse(systemd.available("foo.service"))

        # systemd >= 231
        with patch.dict(systemd.__context__, {"salt.utils.systemd.version": 231}):
            with patch.dict(_SYSTEMCTL_STATUS, _SYSTEMCTL_STATUS_GTE_231):
                with patch.object(systemd, "_systemctl_status", mock), patch.object(
                    systemd, "offline", MagicMock(return_value=False)
                ):
                    self.assertTrue(systemd.available("sshd.service"))
                    self.assertFalse(systemd.available("bar.service"))

        # systemd < 231 with retcode/output changes backported (e.g. RHEL 7.3)
        with patch.dict(systemd.__context__, {"salt.utils.systemd.version": 219}):
            with patch.dict(_SYSTEMCTL_STATUS, _SYSTEMCTL_STATUS_GTE_231):
                with patch.object(systemd, "_systemctl_status", mock), patch.object(
                    systemd, "offline", MagicMock(return_value=False)
                ):
                    self.assertTrue(systemd.available("sshd.service"))
                    self.assertFalse(systemd.available("bar.service"))

    def test_missing(self):
        """
        Test to the inverse of service.available.
        """
        mock = MagicMock(side_effect=lambda x: _SYSTEMCTL_STATUS[x])

        # systemd < 231
        with patch.dict(systemd.__context__, {"salt.utils.systemd.version": 230}):
            with patch.object(systemd, "_systemctl_status", mock), patch.object(
                systemd, "offline", MagicMock(return_value=False)
            ):
                self.assertFalse(systemd.missing("sshd.service"))
                self.assertTrue(systemd.missing("foo.service"))

        # systemd >= 231
        with patch.dict(systemd.__context__, {"salt.utils.systemd.version": 231}):
            with patch.dict(_SYSTEMCTL_STATUS, _SYSTEMCTL_STATUS_GTE_231):
                with patch.object(systemd, "_systemctl_status", mock), patch.object(
                    systemd, "offline", MagicMock(return_value=False)
                ):
                    self.assertFalse(systemd.missing("sshd.service"))
                    self.assertTrue(systemd.missing("bar.service"))

        # systemd < 231 with retcode/output changes backported (e.g. RHEL 7.3)
        with patch.dict(systemd.__context__, {"salt.utils.systemd.version": 219}):
            with patch.dict(_SYSTEMCTL_STATUS, _SYSTEMCTL_STATUS_GTE_231):
                with patch.object(systemd, "_systemctl_status", mock), patch.object(
                    systemd, "offline", MagicMock(return_value=False)
                ):
                    self.assertFalse(systemd.missing("sshd.service"))
                    self.assertTrue(systemd.missing("bar.service"))

    def test_show(self):
        """
        Test to show properties of one or more units/jobs or the manager
        """
        show_output = "a=b\nc=d\ne={ f=g ; h=i }\nWants=foo.service bar.service\n"
        mock = MagicMock(return_value=show_output)
        with patch.dict(systemd.__salt__, {"cmd.run": mock}):
            self.assertDictEqual(
                systemd.show("sshd"),
                {
                    "a": "b",
                    "c": "d",
                    "e": {"f": "g", "h": "i"},
                    "Wants": ["foo.service", "bar.service"],
                },
            )

    def test_execs(self):
        """
        Test to return a list of all files specified as ``ExecStart`` for all
        services
        """
        mock = MagicMock(return_value=["a", "b"])
        with patch.object(systemd, "get_all", mock):
            mock = MagicMock(return_value={"ExecStart": {"path": "c"}})
            with patch.object(systemd, "show", mock):
                self.assertDictEqual(systemd.execs(), {"a": "c", "b": "c"})

    def _check_automatic_daemon_reload(self, reload_needed):
        """
        Test if a needed 'systemctl daemon-reload' is detected and executed when starting a service
        """
        status = _SYSTEMCTL_STATUS["sshd.service"]
        if reload_needed:
            status[
                "stdout"
            ] = f"""\
Warning: The unit file, source configuration file or drop-ins of sshd.service changed on disk. Run 'systemctl daemon-reload' to reload units.
{status["stdout"]}"""

        mock_systemctl_reload = MagicMock(return_value=True)
        mock_run = MagicMock(
            side_effect=lambda x, **_: (
                status
                if "status" in x
                else {"stdout": "", "stderr": "", "retcode": 0, "pid": 54321}
            )
        )

        with patch("salt.utils.path.which", lambda x: "/bin/" + x):
            with patch.object(
                systemd, "_untracked_custom_unit_found", MagicMock(return_value=False)
            ):
                with patch.object(systemd, "systemctl_reload", mock_systemctl_reload):
                    with patch.object(
                        systemd, "_check_unmask", MagicMock(return_value=None)
                    ):
                        with patch.dict(
                            systemd.__salt__,
                            {
                                "config.get": MagicMock(return_value=False),
                                "cmd.run_all": mock_run,
                            },
                        ):
                            self.assertTrue(systemd.start("sshd.service"))
                            self.assertTrue(
                                (mock_systemctl_reload.call_count == 1) == reload_needed
                            )

    def test_automatic_daemon_reload(self):
        self._check_automatic_daemon_reload(False)
        self._check_automatic_daemon_reload(True)


class SystemdScopeTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test case for salt.modules.systemd, for functions which use systemd
    scopes
    """

    def setup_loader_modules(self):
        return {systemd: {}}

    unit_name = "foo"
    mock_none = MagicMock(return_value=None)
    mock_success = MagicMock(return_value=0)
    mock_failure = MagicMock(return_value=1)
    mock_true = MagicMock(return_value=True)
    mock_false = MagicMock(return_value=False)
    mock_empty_list = MagicMock(return_value=[])
    mock_run_all_success = MagicMock(
        return_value={"retcode": 0, "stdout": "", "stderr": "", "pid": 12345}
    )
    mock_run_all_failure = MagicMock(
        return_value={"retcode": 1, "stdout": "", "stderr": "", "pid": 12345}
    )

    def _change_state(self, action, no_block=False):
        """
        Common code for start/stop/restart/reload/force_reload tests
        """
        # We want the traceback if the function name can't be found in the
        # systemd execution module.
        func = getattr(systemd, action)
        # Remove trailing _ in "reload_"
        action = action.rstrip("_").replace("_", "-")
        systemctl_command = ["/bin/systemctl"]
        if no_block:
            systemctl_command.append("--no-block")
        systemctl_command.extend([action, self.unit_name + ".service"])
        scope_prefix = ["/bin/systemd-run", "--scope"]

        assert_kwargs = {"python_shell": False}
        if action in ("enable", "disable"):
            assert_kwargs["ignore_retcode"] = True

        with patch("salt.utils.path.which", lambda x: "/bin/" + x):
            with patch.object(systemd, "_check_for_unit_changes", self.mock_none):
                with patch.object(systemd, "_unit_file_changed", self.mock_none):
                    with patch.object(systemd, "_check_unmask", self.mock_none):
                        with patch.object(
                            systemd, "_get_sysv_services", self.mock_empty_list
                        ):

                            # Has scopes available
                            with patch.object(
                                salt.utils.systemd, "has_scope", self.mock_true
                            ):

                                # Scope enabled, successful
                                with patch.dict(
                                    systemd.__salt__,
                                    {
                                        "config.get": self.mock_true,
                                        "cmd.run_all": self.mock_run_all_success,
                                    },
                                ):
                                    ret = func(self.unit_name, no_block=no_block)
                                    self.assertTrue(ret)
                                    self.mock_run_all_success.assert_called_with(
                                        scope_prefix + systemctl_command,
                                        **assert_kwargs,
                                    )

                                # Scope enabled, failed
                                with patch.dict(
                                    systemd.__salt__,
                                    {
                                        "config.get": self.mock_true,
                                        "cmd.run_all": self.mock_run_all_failure,
                                    },
                                ):
                                    if action in ("stop", "disable"):
                                        ret = func(self.unit_name, no_block=no_block)
                                        self.assertFalse(ret)
                                    else:
                                        self.assertRaises(
                                            CommandExecutionError,
                                            func,
                                            self.unit_name,
                                            no_block=no_block,
                                        )
                                    self.mock_run_all_failure.assert_called_with(
                                        scope_prefix + systemctl_command,
                                        **assert_kwargs,
                                    )

                                # Scope disabled, successful
                                with patch.dict(
                                    systemd.__salt__,
                                    {
                                        "config.get": self.mock_false,
                                        "cmd.run_all": self.mock_run_all_success,
                                    },
                                ):
                                    ret = func(self.unit_name, no_block=no_block)
                                    self.assertTrue(ret)
                                    self.mock_run_all_success.assert_called_with(
                                        systemctl_command, **assert_kwargs
                                    )

                                # Scope disabled, failed
                                with patch.dict(
                                    systemd.__salt__,
                                    {
                                        "config.get": self.mock_false,
                                        "cmd.run_all": self.mock_run_all_failure,
                                    },
                                ):
                                    if action in ("stop", "disable"):
                                        ret = func(self.unit_name, no_block=no_block)
                                        self.assertFalse(ret)
                                    else:
                                        self.assertRaises(
                                            CommandExecutionError,
                                            func,
                                            self.unit_name,
                                            no_block=no_block,
                                        )
                                    self.mock_run_all_failure.assert_called_with(
                                        systemctl_command, **assert_kwargs
                                    )

                            # Does not have scopes available
                            with patch.object(
                                salt.utils.systemd, "has_scope", self.mock_false
                            ):

                                # The results should be the same irrespective of
                                # whether or not scope is enabled, since scope is not
                                # available, so we repeat the below tests with it both
                                # enabled and disabled.
                                for scope_mock in (self.mock_true, self.mock_false):

                                    # Successful
                                    with patch.dict(
                                        systemd.__salt__,
                                        {
                                            "config.get": scope_mock,
                                            "cmd.run_all": self.mock_run_all_success,
                                        },
                                    ):
                                        ret = func(self.unit_name, no_block=no_block)
                                        self.assertTrue(ret)
                                        self.mock_run_all_success.assert_called_with(
                                            systemctl_command, **assert_kwargs
                                        )

                                    # Failed
                                    with patch.dict(
                                        systemd.__salt__,
                                        {
                                            "config.get": scope_mock,
                                            "cmd.run_all": self.mock_run_all_failure,
                                        },
                                    ):
                                        if action in ("stop", "disable"):
                                            ret = func(
                                                self.unit_name, no_block=no_block
                                            )
                                            self.assertFalse(ret)
                                        else:
                                            self.assertRaises(
                                                CommandExecutionError,
                                                func,
                                                self.unit_name,
                                                no_block=no_block,
                                            )
                                        self.mock_run_all_failure.assert_called_with(
                                            systemctl_command, **assert_kwargs
                                        )

    def _mask_unmask(self, action, runtime):
        """
        Common code for mask/unmask tests
        """
        # We want the traceback if the function name can't be found in the
        # systemd execution module, so don't provide a fallback value for the
        # call to getattr() here.
        func = getattr(systemd, action)
        # Remove trailing _ in "unmask_"
        action = action.rstrip("_").replace("_", "-")
        systemctl_command = ["/bin/systemctl", action]
        if runtime:
            systemctl_command.append("--runtime")
        systemctl_command.append(self.unit_name + ".service")
        scope_prefix = ["/bin/systemd-run", "--scope"]

        args = [self.unit_name, runtime]

        masked_mock = self.mock_true if action == "unmask" else self.mock_false

        with patch("salt.utils.path.which", lambda x: "/bin/" + x):
            with patch.object(systemd, "_check_for_unit_changes", self.mock_none):
                if action == "unmask":
                    mock_not_run = MagicMock(
                        return_value={
                            "retcode": 0,
                            "stdout": "",
                            "stderr": "",
                            "pid": 12345,
                        }
                    )
                    with patch.dict(systemd.__salt__, {"cmd.run_all": mock_not_run}):
                        with patch.object(systemd, "masked", self.mock_false):
                            # Test not masked (should take no action and return True)
                            self.assertTrue(systemd.unmask_(self.unit_name))
                            # Also should not have called cmd.run_all
                            self.assertTrue(mock_not_run.call_count == 0)

                with patch.object(systemd, "masked", masked_mock):

                    # Has scopes available
                    with patch.object(salt.utils.systemd, "has_scope", self.mock_true):

                        # Scope enabled, successful
                        with patch.dict(
                            systemd.__salt__,
                            {
                                "config.get": self.mock_true,
                                "cmd.run_all": self.mock_run_all_success,
                            },
                        ):
                            ret = func(*args)
                            self.assertTrue(ret)
                            self.mock_run_all_success.assert_called_with(
                                scope_prefix + systemctl_command,
                                python_shell=False,
                                redirect_stderr=True,
                            )

                        # Scope enabled, failed
                        with patch.dict(
                            systemd.__salt__,
                            {
                                "config.get": self.mock_true,
                                "cmd.run_all": self.mock_run_all_failure,
                            },
                        ):
                            self.assertRaises(CommandExecutionError, func, *args)
                            self.mock_run_all_failure.assert_called_with(
                                scope_prefix + systemctl_command,
                                python_shell=False,
                                redirect_stderr=True,
                            )

                        # Scope disabled, successful
                        with patch.dict(
                            systemd.__salt__,
                            {
                                "config.get": self.mock_false,
                                "cmd.run_all": self.mock_run_all_success,
                            },
                        ):
                            ret = func(*args)
                            self.assertTrue(ret)
                            self.mock_run_all_success.assert_called_with(
                                systemctl_command,
                                python_shell=False,
                                redirect_stderr=True,
                            )

                        # Scope disabled, failed
                        with patch.dict(
                            systemd.__salt__,
                            {
                                "config.get": self.mock_false,
                                "cmd.run_all": self.mock_run_all_failure,
                            },
                        ):
                            self.assertRaises(CommandExecutionError, func, *args)
                            self.mock_run_all_failure.assert_called_with(
                                systemctl_command,
                                python_shell=False,
                                redirect_stderr=True,
                            )

                    # Does not have scopes available
                    with patch.object(salt.utils.systemd, "has_scope", self.mock_false):

                        # The results should be the same irrespective of
                        # whether or not scope is enabled, since scope is not
                        # available, so we repeat the below tests with it both
                        # enabled and disabled.
                        for scope_mock in (self.mock_true, self.mock_false):

                            # Successful
                            with patch.dict(
                                systemd.__salt__,
                                {
                                    "config.get": scope_mock,
                                    "cmd.run_all": self.mock_run_all_success,
                                },
                            ):
                                ret = func(*args)
                                self.assertTrue(ret)
                                self.mock_run_all_success.assert_called_with(
                                    systemctl_command,
                                    python_shell=False,
                                    redirect_stderr=True,
                                )

                            # Failed
                            with patch.dict(
                                systemd.__salt__,
                                {
                                    "config.get": scope_mock,
                                    "cmd.run_all": self.mock_run_all_failure,
                                },
                            ):
                                self.assertRaises(CommandExecutionError, func, *args)
                                self.mock_run_all_failure.assert_called_with(
                                    systemctl_command,
                                    python_shell=False,
                                    redirect_stderr=True,
                                )

    def test_start(self):
        self._change_state("start", no_block=False)
        self._change_state("start", no_block=True)

    def test_stop(self):
        self._change_state("stop", no_block=False)
        self._change_state("stop", no_block=True)

    def test_restart(self):
        self._change_state("restart", no_block=False)
        self._change_state("restart", no_block=True)

    def test_reload(self):
        self._change_state("reload_", no_block=False)
        self._change_state("reload_", no_block=True)

    def test_force_reload(self):
        self._change_state("force_reload", no_block=False)
        self._change_state("force_reload", no_block=True)

    def test_enable(self):
        self._change_state("enable", no_block=False)
        self._change_state("enable", no_block=True)

    def test_disable(self):
        self._change_state("disable", no_block=False)
        self._change_state("disable", no_block=True)

    def test_mask(self):
        self._mask_unmask("mask", False)

    def test_mask_runtime(self):
        self._mask_unmask("mask", True)

    def test_unmask(self):
        self._mask_unmask("unmask_", False)

    def test_unmask_runtime(self):
        self._mask_unmask("unmask_", True)

    def test_firstboot(self):
        """
        Test service.firstboot without parameters
        """
        result = {"retcode": 0, "stdout": "stdout"}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch("salt.utils.path.which", lambda x: "/bin/" + x):
            with patch.dict(systemd.__salt__, salt_mock):
                assert systemd.firstboot()
                salt_mock["cmd.run_all"].assert_called_with(["/bin/systemd-firstboot"])

    def test_firstboot_params(self):
        """
        Test service.firstboot with parameters
        """
        result = {"retcode": 0, "stdout": "stdout"}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch("salt.utils.path.which", lambda x: "/bin/" + x):
            with patch.dict(systemd.__salt__, salt_mock):
                assert systemd.firstboot(
                    locale="en_US.UTF-8",
                    locale_message="en_US.UTF-8",
                    keymap="jp",
                    timezone="Europe/Berlin",
                    hostname="node-001",
                    machine_id="1234567890abcdef",
                    root="/mnt",
                )
                salt_mock["cmd.run_all"].assert_called_with(
                    [
                        "/bin/systemd-firstboot",
                        "--locale",
                        "en_US.UTF-8",
                        "--locale-message",
                        "en_US.UTF-8",
                        "--keymap",
                        "jp",
                        "--timezone",
                        "Europe/Berlin",
                        "--hostname",
                        "node-001",
                        "--machine-ID",
                        "1234567890abcdef",
                        "--root",
                        "/mnt",
                    ]
                )

    def test_firstboot_error(self):
        """
        Test service.firstboot error
        """
        result = {"retcode": 1, "stderr": "error"}
        salt_mock = {
            "cmd.run_all": MagicMock(return_value=result),
        }
        with patch.dict(systemd.__salt__, salt_mock):
            with pytest.raises(CommandExecutionError):
                assert systemd.firstboot()



import logging

import pytest

try:
    import dateutil.parser

    HAS_DATEUTIL_PARSER = True
except ImportError:
    HAS_DATEUTIL_PARSER = False

log = logging.getLogger(__name__)


pytestmark = pytest.mark.skipif(
    HAS_DATEUTIL_PARSER is False,
    reason="The 'dateutil.parser' library is not available",
)


@pytest.mark.slow_test
def test_skip(schedule):
    """
    verify that scheduled job is skipped at the specified time
    """
    job_name = "test_skip"
    job = {
        "schedule": {
            job_name: {
                "function": "test.ping",
                "when": ["11/29/2017 4pm", "11/29/2017 5pm"],
            }
        }
    }

    # Add job to schedule
    schedule.opts.update(job)

    run_time = dateutil.parser.parse("11/29/2017 4:00pm")
    schedule.skip_job(
        job_name,
        {
            "time": run_time.strftime("%Y-%m-%dT%H:%M:%S"),
            "time_fmt": "%Y-%m-%dT%H:%M:%S",
        },
    )

    # Run 11/29/2017 at 4pm
    schedule.eval(now=run_time)
    ret = schedule.job_status(job_name)
    assert "_last_run" not in ret
    assert ret["_skip_reason"] == "skip_explicit"
    assert ret["_skipped_time"] == run_time

    # Run 11/29/2017 at 5pm
    run_time = dateutil.parser.parse("11/29/2017 5:00pm")
    schedule.eval(now=run_time)
    ret = schedule.job_status(job_name)
    assert ret["_last_run"] == run_time


@pytest.mark.slow_test
def test_skip_during_range(schedule):
    """
    verify that scheduled job is skipped during the specified range
    """
    job_name = "test_skip_during_range"
    job = {
        "schedule": {
            job_name: {
                "function": "test.ping",
                "hours": "1",
                "skip_during_range": {
                    "start": "11/29/2017 2pm",
                    "end": "11/29/2017 3pm",
                },
            }
        }
    }

    # Add job to schedule
    schedule.opts.update(job)

    # eval at 1:30pm to prime.
    run_time = dateutil.parser.parse("11/29/2017 1:30pm")
    schedule.eval(now=run_time)
    ret = schedule.job_status(job_name)

    # eval at 2:30pm, will not run during range.
    run_time = dateutil.parser.parse("11/29/2017 2:30pm")
    schedule.eval(now=run_time)
    ret = schedule.job_status(job_name)
    assert "_last_run" not in ret
    assert ret["_skip_reason"] == "in_skip_range"
    assert ret["_skipped_time"] == run_time

    # eval at 3:30pm, will run.
    run_time = dateutil.parser.parse("11/29/2017 3:30pm")
    schedule.eval(now=run_time)
    ret = schedule.job_status(job_name)
    assert ret["_last_run"] == run_time


def test_skip_during_range_invalid_datestring(schedule):
    """
    verify that scheduled job is not not and returns the right error string
    """
    run_time = dateutil.parser.parse("11/29/2017 2:30pm")

    job_name1 = "skip_during_range_invalid_datestring1"
    job1 = {
        "schedule": {
            job_name1: {
                "function": "test.ping",
                "hours": "1",
                "_next_fire_time": run_time,
                "skip_during_range": {"start": "25pm", "end": "3pm"},
            }
        }
    }

    job_name2 = "skip_during_range_invalid_datestring2"
    job2 = {
        "schedule": {
            job_name2: {
                "function": "test.ping",
                "hours": "1",
                "_next_fire_time": run_time,
                "skip_during_range": {"start": "2pm", "end": "25pm"},
            }
        }
    }

    # Add job1 to schedule
    schedule.opts.update(job1)

    # Eval
    schedule.eval(now=run_time)

    # Check the first job
    ret = schedule.job_status(job_name1)
    _expected = (
        "Invalid date string for start in skip_during_range. Ignoring job {}."
    ).format(job_name1)
    assert ret["_error"] == _expected

    # Clear out schedule
    schedule.opts["schedule"] = {}

    # Add job2 to schedule
    schedule.opts.update(job2)

    # Eval
    schedule.eval(now=run_time)

    # Check the second job
    ret = schedule.job_status(job_name2)
    _expected = (
        "Invalid date string for end in skip_during_range. Ignoring job {}."
    ).format(job_name2)
    assert ret["_error"] == _expected


@pytest.mark.slow_test
def test_skip_during_range_global(schedule):
    """
    verify that scheduled job is skipped during the specified range
    """
    job_name = "skip_during_range_global"
    job = {
        "schedule": {
            "skip_during_range": {
                "start": "11/29/2017 2:00pm",
                "end": "11/29/2017 3:00pm",
            },
            job_name: {"function": "test.ping", "hours": "1"},
        }
    }

    # Add job to schedule
    schedule.opts.update(job)

    # eval at 1:30pm to prime.
    run_time = dateutil.parser.parse("11/29/2017 1:30pm")
    schedule.eval(now=run_time)
    ret = schedule.job_status(job_name)

    # eval at 2:30pm, will not run during range.
    run_time = dateutil.parser.parse("11/29/2017 2:30pm")
    schedule.eval(now=run_time)
    ret = schedule.job_status(job_name)
    assert "_last_run" not in ret
    assert ret["_skip_reason"] == "in_skip_range"
    assert ret["_skipped_time"] == run_time

    # eval at 3:30pm, will run.
    run_time = dateutil.parser.parse("11/29/2017 3:30pm")
    schedule.eval(now=run_time)
    ret = schedule.job_status(job_name)
    assert ret["_last_run"] == run_time


@pytest.mark.slow_test
def test_run_after_skip_range(schedule):
    """
    verify that scheduled job is skipped during the specified range
    """
    job_name = "skip_run_after_skip_range"
    job = {
        "schedule": {
            job_name: {
                "function": "test.ping",
                "when": "11/29/2017 2:30pm",
                "run_after_skip_range": True,
                "skip_during_range": {
                    "start": "11/29/2017 2pm",
                    "end": "11/29/2017 3pm",
                },
            }
        }
    }

    # Add job to schedule
    schedule.opts.update(job)

    # eval at 2:30pm, will not run during range.
    run_time = dateutil.parser.parse("11/29/2017 2:30pm")
    schedule.eval(now=run_time)
    ret = schedule.job_status(job_name)
    assert "_last_run" not in ret
    assert ret["_skip_reason"] == "in_skip_range"
    assert ret["_skipped_time"] == run_time

    # eval at 3:00:01pm, will run.
    run_time = dateutil.parser.parse("11/29/2017 3:00:01pm")
    schedule.eval(now=run_time)
    ret = schedule.job_status(job_name)
    assert ret["_last_run"] == run_time


@pytest.mark.slow_test
def test_run_seconds_skip(schedule):
    """
    verify that scheduled job is skipped during the specified range
    """
    job_name = "run_seconds_skip"
    job = {"schedule": {job_name: {"function": "test.ping", "seconds": "10"}}}

    # Add job to schedule
    schedule.opts.update(job)

    # eval at 2:00pm, to prime the scheduler
    run_time = dateutil.parser.parse("11/29/2017 2:00pm")
    schedule.eval(now=run_time)
    ret = schedule.job_status(job_name)

    # eval at 2:00:10pm
    run_time = dateutil.parser.parse("11/29/2017 2:00:10pm")
    schedule.eval(now=run_time)
    ret = schedule.job_status(job_name)

    # Skip at 2:00:20pm
    run_time = dateutil.parser.parse("11/29/2017 2:00:20pm")
    schedule.skip_job(
        job_name,
        {
            "time": run_time.strftime("%Y-%m-%dT%H:%M:%S"),
            "time_fmt": "%Y-%m-%dT%H:%M:%S",
        },
    )
    schedule.eval(now=run_time)
    ret = schedule.job_status(job_name)
    assert "_next_fire_time" in ret
    assert ret["_skip_reason"] == "skip_explicit"
    assert ret["_skipped_time"] == run_time

    # Run at 2:00:30pm
    run_time = dateutil.parser.parse("11/29/2017 2:00:30pm")
    schedule.eval(now=run_time)
    ret = schedule.job_status(job_name)
    assert "_last_run" in ret



"""
Control the OpenBSD packet filter (PF).

:codeauthor: Jasper Lievisse Adriaanse <j@jasper.la>

.. versionadded:: 2019.2.0
"""

import logging
import re

import salt.utils.path
from salt.exceptions import CommandExecutionError, SaltInvocationError

log = logging.getLogger(__name__)


def __virtual__():
    """
    Only works on OpenBSD and FreeBSD for now; other systems with pf (macOS,
    FreeBSD, etc) need to be tested before enabling them.
    """
    tested_oses = ["FreeBSD", "OpenBSD"]
    if __grains__["os"] in tested_oses and salt.utils.path.which("pfctl"):
        return True

    return (
        False,
        "The pf execution module cannot be loaded: either the OS ({}) is not "
        "tested or the pfctl binary was not found".format(__grains__["os"]),
    )


def enable():
    """
    Enable the Packet Filter.

    CLI Example:

    .. code-block:: bash

        salt '*' pf.enable
    """
    ret = {}
    result = __salt__["cmd.run_all"](
        "pfctl -e", output_loglevel="trace", python_shell=False
    )

    if result["retcode"] == 0:
        ret = {"comment": "pf enabled", "changes": True}
    else:
        # If pf was already enabled the return code is also non-zero.
        # Don't raise an exception in that case.
        if result["stderr"] == "pfctl: pf already enabled":
            ret = {"comment": "pf already enabled", "changes": False}
        else:
            raise CommandExecutionError(
                "Could not enable pf",
                info={"errors": [result["stderr"]], "changes": False},
            )

    return ret


def disable():
    """
    Disable the Packet Filter.

    CLI Example:

    .. code-block:: bash

        salt '*' pf.disable
    """
    ret = {}
    result = __salt__["cmd.run_all"](
        "pfctl -d", output_loglevel="trace", python_shell=False
    )

    if result["retcode"] == 0:
        ret = {"comment": "pf disabled", "changes": True}
    else:
        # If pf was already disabled the return code is also non-zero.
        # Don't raise an exception in that case.
        if result["stderr"] == "pfctl: pf not enabled":
            ret = {"comment": "pf already disabled", "changes": False}
        else:
            raise CommandExecutionError(
                "Could not disable pf",
                info={"errors": [result["stderr"]], "changes": False},
            )

    return ret


def loglevel(level):
    """
    Set the debug level which limits the severity of log messages printed by ``pf(4)``.

    level:
        Log level. Should be one of the following: emerg, alert, crit, err, warning, notice,
        info or debug (OpenBSD); or none, urgent, misc, loud (FreeBSD).

    CLI Example:

    .. code-block:: bash

        salt '*' pf.loglevel emerg
    """
    # There's no way to getting the previous loglevel so imply we've
    # always made a change.
    ret = {"changes": True}

    myos = __grains__["os"]
    if myos == "FreeBSD":
        all_levels = ["none", "urgent", "misc", "loud"]
    else:
        all_levels = [
            "emerg",
            "alert",
            "crit",
            "err",
            "warning",
            "notice",
            "info",
            "debug",
        ]
    if level not in all_levels:
        raise SaltInvocationError(f"Unknown loglevel: {level}")

    result = __salt__["cmd.run_all"](
        f"pfctl -x {level}", output_loglevel="trace", python_shell=False
    )

    if result["retcode"] != 0:
        raise CommandExecutionError(
            "Problem encountered setting loglevel",
            info={"errors": [result["stderr"]], "changes": False},
        )

    return ret


def load(file="/etc/pf.conf", noop=False):
    """
    Load a ruleset from the specific file, overwriting the currently loaded ruleset.

    file:
        Full path to the file containing the ruleset.

    noop:
        Don't actually load the rules, just parse them.

    CLI Example:

    .. code-block:: bash

        salt '*' pf.load /etc/pf.conf.d/lockdown.conf
    """
    # We cannot precisely determine if loading the ruleset implied
    # any changes so assume it always does.
    ret = {"changes": True}
    cmd = ["pfctl", "-f", file]

    if noop:
        ret["changes"] = False
        cmd.append("-n")

    result = __salt__["cmd.run_all"](cmd, output_loglevel="trace", python_shell=False)

    if result["retcode"] != 0:
        raise CommandExecutionError(
            f"Problem loading the ruleset from {file}",
            info={"errors": [result["stderr"]], "changes": False},
        )

    return ret


def flush(modifier):
    """
    Flush the specified packet filter parameters.

    modifier:
        Should be one of the following:

        - all
        - info
        - osfp
        - rules
        - sources
        - states
        - tables

        Please refer to the OpenBSD `pfctl(8) <https://man.openbsd.org/pfctl#T>`_
        documentation for a detailed explanation of each command.

    CLI Example:

    .. code-block:: bash

        salt '*' pf.flush states
    """
    ret = {}

    all_modifiers = ["rules", "states", "info", "osfp", "all", "sources", "tables"]

    # Accept the following two modifiers to allow for a consistent interface between
    # pfctl(8) and Salt.
    capital_modifiers = ["Sources", "Tables"]
    all_modifiers += capital_modifiers
    if modifier.title() in capital_modifiers:
        modifier = modifier.title()

    if modifier not in all_modifiers:
        raise SaltInvocationError(f"Unknown modifier: {modifier}")

    cmd = f"pfctl -v -F {modifier}"
    result = __salt__["cmd.run_all"](cmd, output_loglevel="trace", python_shell=False)

    if result["retcode"] == 0:
        if re.match(r"^0.*", result["stderr"]):
            ret["changes"] = False
        else:
            ret["changes"] = True

        ret["comment"] = result["stderr"]
    else:
        raise CommandExecutionError(
            f"Could not flush {modifier}",
            info={"errors": [result["stderr"]], "changes": False},
        )

    return ret


def table(command, table, **kwargs):
    """
    Apply a command on the specified table.

    table:
        Name of the table.

    command:
        Command to apply to the table. Supported commands are:

        - add
        - delete
        - expire
        - flush
        - kill
        - replace
        - show
        - test
        - zero

        Please refer to the OpenBSD `pfctl(8) <https://man.openbsd.org/pfctl#T>`_
        documentation for a detailed explanation of each command.

    CLI Example:

    .. code-block:: bash

        salt '*' pf.table expire table=spam_hosts number=300
        salt '*' pf.table add table=local_hosts addresses='["127.0.0.1", "::1"]'
    """
    ret = {}

    all_commands = [
        "kill",
        "flush",
        "add",
        "delete",
        "expire",
        "replace",
        "show",
        "test",
        "zero",
    ]
    if command not in all_commands:
        raise SaltInvocationError(f"Unknown table command: {command}")

    cmd = ["pfctl", "-t", table, "-T", command]

    if command in ["add", "delete", "replace", "test"]:
        cmd += kwargs.get("addresses", [])
    elif command == "expire":
        number = kwargs.get("number", None)
        if not number:
            raise SaltInvocationError("need expire_number argument for expire command")
        else:
            cmd.append(number)

    result = __salt__["cmd.run_all"](cmd, output_level="trace", python_shell=False)

    if result["retcode"] == 0:
        if command == "show":
            ret = {"comment": result["stdout"].split()}
        elif command == "test":
            ret = {"comment": result["stderr"], "matches": True}
        else:
            if re.match(r"^(0.*|no changes)", result["stderr"]):
                ret["changes"] = False
            else:
                ret["changes"] = True

            ret["comment"] = result["stderr"]
    else:
        # 'test' returns a non-zero code if the address didn't match, even if
        # the command itself ran fine; also set 'matches' to False since not
        # everything matched.
        if command == "test" and re.match(
            r"^\d+/\d+ addresses match.$", result["stderr"]
        ):
            ret = {"comment": result["stderr"], "matches": False}
        else:
            raise CommandExecutionError(
                f"Could not apply {command} on table {table}",
                info={"errors": [result["stderr"]], "changes": False},
            )

    return ret


def show(modifier):
    """
    Show filter parameters.

    modifier:
        Modifier to apply for filtering. Only a useful subset of what pfctl supports
        can be used with Salt.

        - rules
        - states
        - tables

    CLI Example:

    .. code-block:: bash

        salt '*' pf.show rules
    """
    # By definition showing the parameters makes no changes.
    ret = {"changes": False}

    capital_modifiers = ["Tables"]
    all_modifiers = ["rules", "states", "tables"]
    all_modifiers += capital_modifiers
    if modifier.title() in capital_modifiers:
        modifier = modifier.title()

    if modifier not in all_modifiers:
        raise SaltInvocationError(f"Unknown modifier: {modifier}")

    cmd = f"pfctl -s {modifier}"
    result = __salt__["cmd.run_all"](cmd, output_loglevel="trace", python_shell=False)

    if result["retcode"] == 0:
        ret["comment"] = result["stdout"].split("\n")
    else:
        raise CommandExecutionError(
            f"Could not show {modifier}",
            info={"errors": [result["stderr"]], "changes": False},
        )

    return ret



"""
Tests for salt.utils.jinja
"""

import pytest


class MockFileClient:
    """
    Does not download files but records any file request for testing
    """

    def __init__(self, loader=None):
        if loader:
            loader._file_client = self
        self.requests = []
        self.opts = {}

    def get_file(self, template, dest="", makedirs=False, saltenv="base"):
        self.requests.append(
            {"path": template, "dest": dest, "makedirs": makedirs, "saltenv": saltenv}
        )


@pytest.fixture
def mock_file_client(loader=None):
    return MockFileClient(loader)


@pytest.fixture
def template_dir(tmp_path):
    templates_dir = tmp_path / "files" / "test"
    templates_dir.mkdir(parents=True, exist_ok=True)
    return templates_dir


@pytest.fixture
def hello_import(macro_template, template_dir):
    contents = """{% from 'macro' import mymacro -%}
    {% from 'macro' import mymacro -%}
    {{ mymacro('Hey') ~ mymacro(a|default('a'), b|default('b')) }}
    """

    with pytest.helpers.temp_file(
        "hello_import", directory=template_dir, contents=contents
    ) as hello_import_filename:
        yield hello_import_filename


@pytest.fixture
def macro_template(template_dir):
    contents = """# macro
    {% macro mymacro(greeting, greetee='world') -%}
    {{ greeting ~ ' ' ~ greetee }} !
    {%- endmacro %}
    """

    with pytest.helpers.temp_file(
        "macro", directory=template_dir, contents=contents
    ) as macro_filename:
        yield macro_filename



import pytest


@pytest.fixture(scope="module")
def state_tree(base_env_state_tree_root_dir):
    top_file = """
    {%- from "map.jinja" import abc with context %}
    base:
      'localhost':
        - basic
      '127.0.0.1':
        - basic
    """
    map_file = """
    {%- set abc = "def" %}
    """
    state_file = """
    {%- from "map.jinja" import abc with context %}
    Ok with {{ abc }}:
      test.succeed_with_changes
    """
    top_tempfile = pytest.helpers.temp_file(
        "top.sls", top_file, base_env_state_tree_root_dir
    )
    map_tempfile = pytest.helpers.temp_file(
        "map.jinja", map_file, base_env_state_tree_root_dir
    )
    state_tempfile = pytest.helpers.temp_file(
        "test.sls", state_file, base_env_state_tree_root_dir
    )
    with top_tempfile, map_tempfile, state_tempfile:
        yield


@pytest.fixture(scope="module")
def state_tree_dir(base_env_state_tree_root_dir):
    """
    State tree with files to test salt-ssh
    when the map.jinja file is in another directory
    """
    # Remove unused import from top file to avoid salt-ssh file sync issues
    # Use "testdir" instead of "test" to avoid conflicts with state_tree fixture
    top_file = """
    base:
      'localhost':
        - testdir
      '127.0.0.1':
        - testdir
    """
    map_file = """
    {%- set abc = "def" %}
    """
    # State file imports from subdirectory - this is what we're testing
    state_file = """
    {%- from "test/map.jinja" import abc with context %}

    Ok with {{ abc }}:
      test.succeed_without_changes
    """
    top_tempfile = pytest.helpers.temp_file(
        "top.sls", top_file, base_env_state_tree_root_dir
    )
    map_tempfile = pytest.helpers.temp_file(
        "test/map.jinja", map_file, base_env_state_tree_root_dir
    )
    # Use testdir.sls to avoid collision with state_tree's test.sls
    state_tempfile = pytest.helpers.temp_file(
        "testdir.sls", state_file, base_env_state_tree_root_dir
    )

    with top_tempfile, map_tempfile, state_tempfile:
        yield


@pytest.fixture
def nested_state_tree(base_env_state_tree_root_dir, tmp_path):
    top_file = """
    base:
      'localhost':
        - basic
      '127.0.0.1':
        - basic
    """
    # Inline the Jinja template content to avoid salt-ssh file server lookup issues
    # The template imports from foo/map.jinja which still needs to exist as a file
    file_jinja_content = """{% from 'foo/map.jinja' import comment %}{{ comment }}"""
    state_file = """
    /{}/file.txt:
      file.managed:
        - contents: |
            {}
        - template: jinja
    """.format(
        tmp_path, file_jinja_content
    )
    map_file = """
    {% set comment = "blah blah" %}
    """
    statedir = base_env_state_tree_root_dir / "foo"
    top_tempfile = pytest.helpers.temp_file(
        "top.sls", top_file, base_env_state_tree_root_dir
    )
    map_tempfile = pytest.helpers.temp_file("map.jinja", map_file, statedir)
    state_tempfile = pytest.helpers.temp_file("init.sls", state_file, statedir)

    with top_tempfile, map_tempfile, state_tempfile:
        yield


@pytest.fixture(scope="module")
def pillar_tree_nested(base_env_pillar_tree_root_dir):
    top_file = """
    base:
      'localhost':
        - nested
      '127.0.0.1':
        - nested
    """
    nested_pillar = r"""
    {%- do salt.log.warning("hithere: pillar was rendered") %}
    monty: python
    the_meaning:
      of:
        life: 42
        bar: tender
      for: what
    """
    top_tempfile = pytest.helpers.temp_file(
        "top.sls", top_file, base_env_pillar_tree_root_dir
    )
    nested_tempfile = pytest.helpers.temp_file(
        "nested.sls", nested_pillar, base_env_pillar_tree_root_dir
    )
    with top_tempfile, nested_tempfile:
        yield



"""
virt.query outputter
====================

Used to display the output from the :mod:`virt.query <salt.runners.virt.query>`
runner.
"""


def output(data, **kwargs):  # pylint: disable=unused-argument
    """
    Display output for the salt-run virt.query function
    """
    out = ""
    for id_ in data["data"]:
        out += f"{id_}\n"
        for vm_ in data["data"][id_]["vm_info"]:
            out += f"  {vm_}\n"
            vm_data = data[id_]["vm_info"][vm_]
            if "cpu" in vm_data:
                out += "    CPU: {}\n".format(vm_data["cpu"])
            if "mem" in vm_data:
                out += "    Memory: {}\n".format(vm_data["mem"])
            if "state" in vm_data:
                out += "    State: {}\n".format(vm_data["state"])
            if "graphics" in vm_data:
                if vm_data["graphics"].get("type", "") == "vnc":
                    out += "    Graphics: vnc - {}:{}\n".format(
                        id_, vm_data["graphics"]["port"]
                    )
            if "disks" in vm_data:
                for disk, d_data in vm_data["disks"].items():
                    out += f"    Disk - {disk}:\n"
                    out += "      Size: {}\n".format(d_data["disk size"])
                    out += "      File: {}\n".format(d_data["file"])
                    out += "      File Format: {}\n".format(d_data["file format"])
            if "nics" in vm_data:
                for mac in vm_data["nics"]:
                    out += f"    Nic - {mac}:\n"
                    out += "      Source: {}\n".format(
                        vm_data["nics"][mac]["source"][
                            next(iter(vm_data["nics"][mac]["source"].keys()))
                        ]
                    )
                    out += "      Type: {}\n".format(vm_data["nics"][mac]["type"])
    return out



"""
    Test parallels desktop execution module functions
"""

import textwrap

import pytest

import salt.modules.parallels as parallels
from salt.exceptions import CommandExecutionError, SaltInvocationError
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {parallels: {}}


def test__normalize_args():
    """
    Test parallels._normalize_args
    """

    def _validate_ret(ret):
        """
        Assert that the returned data is a list of strings
        """
        assert isinstance(ret, list)
        for arg in ret:
            assert isinstance(arg, str)

    # Validate string arguments
    str_args = (
        "electrolytes --aqueous --anion hydroxide --cation=ammonium free radicals"
        " -- hydrogen"
    )
    _validate_ret(parallels._normalize_args(str_args))

    # Validate list arguments
    list_args = " ".join(str_args)
    _validate_ret(parallels._normalize_args(list_args))

    # Validate tuple arguments
    tuple_args = tuple(list_args)
    _validate_ret(parallels._normalize_args(tuple_args))

    # Validate dictionary arguments
    other_args = {"anion": "hydroxide", "cation": "ammonium"}
    _validate_ret(parallels._normalize_args(other_args))


def test__find_guids():
    """
    Test parallels._find_guids
    """
    guid_str = textwrap.dedent(
        """
        PARENT_SNAPSHOT_ID                      SNAPSHOT_ID
                                                {a5b8999f-5d95-4aff-82de-e515b0101b66}
        {a5b8999f-5d95-4aff-82de-e515b0101b66} *{a7345be5-ab66-478c-946e-a6c2caf14909}
    """
    )
    guids = [
        "a5b8999f-5d95-4aff-82de-e515b0101b66",
        "a7345be5-ab66-478c-946e-a6c2caf14909",
    ]

    assert parallels._find_guids(guid_str) == guids


def test_prlsrvctl():
    """
    Test parallels.prlsrvctl
    """
    runas = "macdev"

    # Test missing prlsrvctl binary
    with patch("salt.utils.path.which", MagicMock(return_value=False)):
        with pytest.raises(CommandExecutionError):
            parallels.prlsrvctl("info", runas=runas)

    # Simulate the existence of prlsrvctl
    with patch("salt.utils.path.which", MagicMock(return_value="/usr/bin/prlsrvctl")):

        # Validate 'prlsrvctl info'
        info_cmd = ["prlsrvctl", "info"]
        info_fcn = MagicMock()
        with patch.dict(parallels.__salt__, {"cmd.run": info_fcn}):
            parallels.prlsrvctl("info", runas=runas)
            info_fcn.assert_called_once_with(info_cmd, runas=runas)

        # Validate 'prlsrvctl usb list'
        usb_cmd = ["prlsrvctl", "usb", "list"]
        usb_fcn = MagicMock()
        with patch.dict(parallels.__salt__, {"cmd.run": usb_fcn}):
            parallels.prlsrvctl("usb", "list", runas=runas)
            usb_fcn.assert_called_once_with(usb_cmd, runas=runas)

        # Validate 'prlsrvctl set "--mem-limit auto"'
        set_cmd = ["prlsrvctl", "set", "--mem-limit", "auto"]
        set_fcn = MagicMock()
        with patch.dict(parallels.__salt__, {"cmd.run": set_fcn}):
            parallels.prlsrvctl("set", "--mem-limit auto", runas=runas)
            set_fcn.assert_called_once_with(set_cmd, runas=runas)


def test_prlctl():
    """
    Test parallels.prlctl
    """
    runas = "macdev"

    # Test missing prlctl binary
    with patch("salt.utils.path.which", MagicMock(return_value=False)):
        with pytest.raises(CommandExecutionError):
            parallels.prlctl("info", runas=runas)

    # Simulate the existence of prlctl
    with patch("salt.utils.path.which", MagicMock(return_value="/usr/bin/prlctl")):

        # Validate 'prlctl user list'
        user_cmd = ["prlctl", "user", "list"]
        user_fcn = MagicMock()
        with patch.dict(parallels.__salt__, {"cmd.run": user_fcn}):
            parallels.prlctl("user", "list", runas=runas)
            user_fcn.assert_called_once_with(user_cmd, runas=runas)

        # Validate 'prlctl exec "macvm uname"'
        exec_cmd = ["prlctl", "exec", "macvm", "uname"]
        exec_fcn = MagicMock()
        with patch.dict(parallels.__salt__, {"cmd.run": exec_fcn}):
            parallels.prlctl("exec", "macvm uname", runas=runas)
            exec_fcn.assert_called_once_with(exec_cmd, runas=runas)

        # Validate 'prlctl capture "macvm --file macvm.display.png"'
        cap_cmd = ["prlctl", "capture", "macvm", "--file", "macvm.display.png"]
        cap_fcn = MagicMock()
        with patch.dict(parallels.__salt__, {"cmd.run": cap_fcn}):
            parallels.prlctl("capture", "macvm --file macvm.display.png", runas=runas)
            cap_fcn.assert_called_once_with(cap_cmd, runas=runas)


def test_list_vms():
    """
    Test parallels.list_vms
    """
    runas = "macdev"

    # Validate a simple list
    mock_plain = MagicMock()
    with patch.object(parallels, "prlctl", mock_plain):
        parallels.list_vms(runas=runas)
        mock_plain.assert_called_once_with("list", [], runas=runas)

    # Validate listing a single VM
    mock_name = MagicMock()
    with patch.object(parallels, "prlctl", mock_name):
        parallels.list_vms(name="macvm", runas=runas)
        mock_name.assert_called_once_with("list", ["macvm"], runas=runas)

    # Validate listing templates
    mock_templ = MagicMock()
    with patch.object(parallels, "prlctl", mock_templ):
        parallels.list_vms(template=True, runas=runas)
        mock_templ.assert_called_once_with("list", ["--template"], runas=runas)

    # Validate listing extra info
    mock_info = MagicMock()
    with patch.object(parallels, "prlctl", mock_info):
        parallels.list_vms(info=True, runas=runas)
        mock_info.assert_called_once_with("list", ["--info"], runas=runas)

    # Validate listing with extra options
    mock_complex = MagicMock()
    with patch.object(parallels, "prlctl", mock_complex):
        parallels.list_vms(args=" -o uuid,status", all=True, runas=runas)
        mock_complex.assert_called_once_with(
            "list", ["-o", "uuid,status", "--all"], runas=runas
        )


def test_clone():
    """
    Test parallels.clone
    """
    name = "macvm"
    runas = "macdev"

    # Validate clone
    mock_clone = MagicMock()
    with patch.object(parallels, "prlctl", mock_clone):
        parallels.clone(name, "macvm_new", runas=runas)
        mock_clone.assert_called_once_with(
            "clone", [name, "--name", "macvm_new"], runas=runas
        )

    # Validate linked clone
    mock_linked = MagicMock()
    with patch.object(parallels, "prlctl", mock_linked):
        parallels.clone(name, "macvm_link", linked=True, runas=runas)
        mock_linked.assert_called_once_with(
            "clone", [name, "--name", "macvm_link", "--linked"], runas=runas
        )

    # Validate template clone
    mock_template = MagicMock()
    with patch.object(parallels, "prlctl", mock_template):
        parallels.clone(name, "macvm_templ", template=True, runas=runas)
        mock_template.assert_called_once_with(
            "clone", [name, "--name", "macvm_templ", "--template"], runas=runas
        )


def test_delete():
    """
    Test parallels.delete
    """
    name = "macvm"
    runas = "macdev"

    # Validate delete
    mock_delete = MagicMock()
    with patch.object(parallels, "prlctl", mock_delete):
        parallels.delete(name, runas=runas)
        mock_delete.assert_called_once_with("delete", name, runas=runas)


def test_exists():
    """
    Test parallels.exists
    """
    name = "macvm"
    runas = "macdev"

    # Validate exists
    mock_list = MagicMock(return_value=f"Name: {name}\nState: running")
    with patch.object(parallels, "list_vms", mock_list):
        assert parallels.exists(name, runas=runas)

    # Validate not exists
    mock_list = MagicMock(return_value=f"Name: {name}\nState: running")
    with patch.object(parallels, "list_vms", mock_list):
        assert not parallels.exists("winvm", runas=runas)


def test_start():
    """
    Test parallels.start
    """
    name = "macvm"
    runas = "macdev"

    mock_start = MagicMock()
    with patch.object(parallels, "prlctl", mock_start):
        parallels.start(name, runas=runas)
        mock_start.assert_called_once_with("start", name, runas=runas)


def test_stop():
    """
    Test parallels.stop
    """
    name = "macvm"
    runas = "macdev"

    # Validate stop
    mock_stop = MagicMock()
    with patch.object(parallels, "prlctl", mock_stop):
        parallels.stop(name, runas=runas)
        mock_stop.assert_called_once_with("stop", [name], runas=runas)

    # Validate immediate stop
    mock_kill = MagicMock()
    with patch.object(parallels, "prlctl", mock_kill):
        parallels.stop(name, kill=True, runas=runas)
        mock_kill.assert_called_once_with("stop", [name, "--kill"], runas=runas)


def test_restart():
    """
    Test parallels.restart
    """
    name = "macvm"
    runas = "macdev"

    mock_start = MagicMock()
    with patch.object(parallels, "prlctl", mock_start):
        parallels.restart(name, runas=runas)
        mock_start.assert_called_once_with("restart", name, runas=runas)


def test_reset():
    """
    Test parallels.reset
    """
    name = "macvm"
    runas = "macdev"

    mock_start = MagicMock()
    with patch.object(parallels, "prlctl", mock_start):
        parallels.reset(name, runas=runas)
        mock_start.assert_called_once_with("reset", name, runas=runas)


def test_status():
    """
    Test parallels.status
    """
    name = "macvm"
    runas = "macdev"

    mock_start = MagicMock()
    with patch.object(parallels, "prlctl", mock_start):
        parallels.status(name, runas=runas)
        mock_start.assert_called_once_with("status", name, runas=runas)


def test_exec_():
    """
    Test parallels.exec_
    """
    name = "macvm"
    runas = "macdev"

    mock_start = MagicMock()
    with patch.object(parallels, "prlctl", mock_start):
        parallels.exec_(name, "find /etc/paths.d", runas=runas)
        mock_start.assert_called_once_with(
            "exec", [name, "find", "/etc/paths.d"], runas=runas
        )


def test_snapshot_id_to_name():
    """
    Test parallels.snapshot_id_to_name
    """
    name = "macvm"
    snap_id = "a5b8999f-5d95-4aff-82de-e515b0101b66"

    # Invalid GUID raises error
    pytest.raises(
        SaltInvocationError, parallels.snapshot_id_to_name, name, "{8-4-4-4-12}"
    )

    # Empty return from prlctl raises error (name/snap_id mismatch?)
    mock_no_data = MagicMock(return_value="")
    with patch.object(parallels, "prlctl", mock_no_data):
        pytest.raises(SaltInvocationError, parallels.snapshot_id_to_name, name, snap_id)

    # Data returned from prlctl is invalid YAML
    mock_invalid_data = MagicMock(return_value="[string theory is falsifiable}")
    with patch.object(parallels, "prlctl", mock_invalid_data):
        snap_name = parallels.snapshot_id_to_name(name, snap_id)
        assert snap_name == ""

    # Data returned from prlctl does not render as a dictionary
    mock_unknown_data = MagicMock(return_value="['sfermions', 'bosinos']")
    with patch.object(parallels, "prlctl", mock_unknown_data):
        snap_name = parallels.snapshot_id_to_name(name, snap_id)
        assert snap_name == ""

    # Snapshot is unnamed
    mock_no_name = MagicMock(return_value="Name:")
    with patch.object(parallels, "prlctl", mock_no_name):
        snap_name = parallels.snapshot_id_to_name(name, snap_id)
        assert snap_name == ""

    # If strict, then raise an error when name is not found
    mock_no_name = MagicMock(return_value="Name:")
    with patch.object(parallels, "prlctl", mock_no_name):
        pytest.raises(
            SaltInvocationError,
            parallels.snapshot_id_to_name,
            name,
            snap_id,
            strict=True,
        )

    # Return name when found
    mock_yes_name = MagicMock(return_value="Name: top")
    with patch.object(parallels, "prlctl", mock_yes_name):
        snap_name = parallels.snapshot_id_to_name(name, snap_id)
        assert snap_name == "top"


def test_snapshot_name_to_id():
    """
    Test parallels.snapshot_name_to_id
    """
    name = "macvm"
    snap_ids = [
        "a5b8999f-5d95-4aff-82de-e515b0101b66",
        "a7345be5-ab66-478c-946e-a6c2caf14909",
    ]
    snap_id = snap_ids[0]
    guid_str = textwrap.dedent(
        """
        PARENT_SNAPSHOT_ID                      SNAPSHOT_ID
                                                {a5b8999f-5d95-4aff-82de-e515b0101b66}
        {a5b8999f-5d95-4aff-82de-e515b0101b66} *{a7345be5-ab66-478c-946e-a6c2caf14909}
    """
    )
    mock_guids = MagicMock(return_value=guid_str)

    # Raise error when no IDs found for snapshot name
    with patch.object(parallels, "prlctl", mock_guids):
        mock_no_names = MagicMock(return_value=[])
        with patch.object(parallels, "snapshot_id_to_name", mock_no_names):
            pytest.raises(
                SaltInvocationError, parallels.snapshot_name_to_id, name, "graviton"
            )

    # Validate singly-valued name
    with patch.object(parallels, "prlctl", mock_guids):
        mock_one_name = MagicMock(side_effect=["", "ν_e"])
        with patch.object(parallels, "snapshot_id_to_name", mock_one_name):
            assert parallels.snapshot_name_to_id(name, "ν_e") == snap_ids[1]

    # Validate multiply-valued name
    with patch.object(parallels, "prlctl", mock_guids):
        mock_many_names = MagicMock(side_effect=["J/Ψ", "J/Ψ"])
        with patch.object(parallels, "snapshot_id_to_name", mock_many_names):
            assert sorted(parallels.snapshot_name_to_id(name, "J/Ψ")) == sorted(
                snap_ids
            )

    # Raise error for multiply-valued name
    with patch.object(parallels, "prlctl", mock_guids):
        mock_many_names = MagicMock(side_effect=["J/Ψ", "J/Ψ"])
        with patch.object(parallels, "snapshot_id_to_name", mock_many_names):
            pytest.raises(
                SaltInvocationError,
                parallels.snapshot_name_to_id,
                name,
                "J/Ψ",
                strict=True,
            )


def test__validate_snap_name():
    """
    Test parallels._validate_snap_name
    """
    name = "macvm"
    snap_id = "a5b8999f-5d95-4aff-82de-e515b0101b66"

    # Validate a GUID passthrough
    assert parallels._validate_snap_name(name, snap_id) == snap_id

    # Validate an unicode name
    mock_snap_symb = MagicMock(return_value=snap_id)
    with patch.object(parallels, "snapshot_name_to_id", mock_snap_symb):
        assert parallels._validate_snap_name(name, "π") == snap_id
        mock_snap_symb.assert_called_once_with(name, "π", strict=True, runas=None)

    # Validate an ascii name
    mock_snap_name = MagicMock(return_value=snap_id)
    with patch.object(parallels, "snapshot_name_to_id", mock_snap_name):
        assert parallels._validate_snap_name(name, "pion") == snap_id
        mock_snap_name.assert_called_once_with(name, "pion", strict=True, runas=None)

    # Validate a numerical name
    mock_snap_numb = MagicMock(return_value=snap_id)
    with patch.object(parallels, "snapshot_name_to_id", mock_snap_numb):
        assert parallels._validate_snap_name(name, "3.14159") == snap_id
        mock_snap_numb.assert_called_once_with(name, "3.14159", strict=True, runas=None)

    # Validate not strict (think neutrino oscillation)
    mock_snap_non_strict = MagicMock(return_value=snap_id)
    with patch.object(parallels, "snapshot_name_to_id", mock_snap_non_strict):
        assert parallels._validate_snap_name(name, "e_ν", strict=False) == snap_id
        mock_snap_non_strict.assert_called_once_with(
            name, "e_ν", strict=False, runas=None
        )


def test_list_snapshots():
    """
    Test parallels.list_snapshots
    """
    name = "macvm"
    guid_str = textwrap.dedent(
        """
        PARENT_SNAPSHOT_ID                      SNAPSHOT_ID
                                                {a5b8999f-5d95-4aff-82de-e515b0101b66}
        {a5b8999f-5d95-4aff-82de-e515b0101b66} *{a7345be5-ab66-478c-946e-a6c2caf14909}
        {a5b8999f-5d95-4aff-82de-e515b0101b66}  {5da9faef-cb0e-466d-9b41-e5571b62ac2a}
    """
    )

    # Validate listing all snapshots for the VM
    mock_prlctl = MagicMock(return_value=guid_str)
    with patch.object(parallels, "prlctl", mock_prlctl):
        parallels.list_snapshots(name)
        mock_prlctl.assert_called_once_with("snapshot-list", [name], runas=None)

    # Validate listing all snapshots in tree mode
    mock_prlctl = MagicMock(return_value=guid_str)
    with patch.object(parallels, "prlctl", mock_prlctl):
        parallels.list_snapshots(name, tree=True)
        mock_prlctl.assert_called_once_with(
            "snapshot-list", [name, "--tree"], runas=None
        )

    # Validate listing a single snapshot
    snap_name = "muon"
    mock_snap_name = MagicMock(return_value=snap_name)
    with patch.object(parallels, "_validate_snap_name", mock_snap_name):
        mock_prlctl = MagicMock(return_value=guid_str)
        with patch.object(parallels, "prlctl", mock_prlctl):
            parallels.list_snapshots(name, snap_name)
            mock_prlctl.assert_called_once_with(
                "snapshot-list", [name, "--id", snap_name], runas=None
            )

    # Validate listing snapshot ID and name pairs
    snap_names = ["electron", "muon", "tauon"]
    mock_snap_name = MagicMock(side_effect=snap_names)
    with patch.object(parallels, "snapshot_id_to_name", mock_snap_name):
        mock_prlctl = MagicMock(return_value=guid_str)
        with patch.object(parallels, "prlctl", mock_prlctl):
            ret = parallels.list_snapshots(name, names=True)
            for snap_name in snap_names:
                assert snap_name in ret
            mock_prlctl.assert_called_once_with("snapshot-list", [name], runas=None)


def test_snapshot():
    """
    Test parallels.snapshot
    """
    name = "macvm"

    # Validate creating a snapshot
    mock_snap = MagicMock(return_value="")
    with patch.object(parallels, "prlctl", mock_snap):
        parallels.snapshot(name)
        mock_snap.assert_called_once_with("snapshot", [name], runas=None)

    # Validate creating a snapshot with a name
    snap_name = "h_0"
    mock_snap_name = MagicMock(return_value="")
    with patch.object(parallels, "prlctl", mock_snap_name):
        parallels.snapshot(name, snap_name)
        mock_snap_name.assert_called_once_with(
            "snapshot", [name, "--name", snap_name], runas=None
        )

    # Validate creating a snapshot with a name and a description
    snap_name = "h_0"
    snap_desc = textwrap.dedent(
        "The ground state particle of the higgs multiplet family of bosons"
    )
    mock_snap_name = MagicMock(return_value="")
    with patch.object(parallels, "prlctl", mock_snap_name):
        parallels.snapshot(name, snap_name, snap_desc)
        mock_snap_name.assert_called_once_with(
            "snapshot",
            [name, "--name", snap_name, "--description", snap_desc],
            runas=None,
        )


def test_delete_snapshot():
    """
    Test parallels.delete_snapshot
    """
    delete_message = (
        "Delete the snapshot...\nThe snapshot has been successfully deleted."
    )

    # Validate single ID
    name = "macvm"
    snap_name = "kaon"
    snap_id = "c2eab062-a635-4ccd-b9ae-998370f898b5"

    mock_snap_name = MagicMock(return_value=snap_id)
    with patch.object(parallels, "_validate_snap_name", mock_snap_name):
        mock_delete = MagicMock(return_value=delete_message)
        with patch.object(parallels, "prlctl", mock_delete):
            ret = parallels.delete_snapshot(name, snap_name)
            assert ret == delete_message
            mock_delete.assert_called_once_with(
                "snapshot-delete", [name, "--id", snap_id], runas=None
            )

    # Validate multiple IDs
    name = "macvm"
    snap_name = "higgs doublet"
    snap_ids = [
        "c2eab062-a635-4ccd-b9ae-998370f898b5",
        "8aca07c5-a0e1-4dcb-ba75-cb154d46d516",
    ]

    mock_snap_ids = MagicMock(return_value=snap_ids)
    with patch.object(parallels, "_validate_snap_name", mock_snap_ids):
        mock_delete = MagicMock(return_value=delete_message)
        with patch.object(parallels, "prlctl", mock_delete):
            ret = parallels.delete_snapshot(name, snap_name, all=True)
            mock_ret = {snap_ids[0]: delete_message, snap_ids[1]: delete_message}
            assert ret == mock_ret
            mock_delete.assert_any_call(
                "snapshot-delete", [name, "--id", snap_ids[0]], runas=None
            )
            mock_delete.assert_any_call(
                "snapshot-delete", [name, "--id", snap_ids[1]], runas=None
            )


def test_revert_snapshot():
    """
    Test parallels.revert_snapshot
    """
    name = "macvm"
    snap_name = "k-bar"
    snap_id = "c2eab062-a635-4ccd-b9ae-998370f898b5"

    mock_snap_name = MagicMock(return_value=snap_id)
    with patch.object(parallels, "_validate_snap_name", mock_snap_name):
        mock_delete = MagicMock(return_value="")
        with patch.object(parallels, "prlctl", mock_delete):
            parallels.revert_snapshot(name, snap_name)
            mock_delete.assert_called_once_with(
                "snapshot-switch", [name, "--id", snap_id], runas=None
            )
